# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import numpy as np
import pyomo.environ as pyo
from pyomo.core.base.block import BlockData, declare_custom_block


def _f_cubic(x, alpha, s=None):
    """
    Cubic function:
        y = a1 + a2 * x + a3 * x^2 + a4 * x^3

    Optionally if s is provided it is a segment index.
        y = a1[s] + a2[s] * x + a3[s] * x^2 + a4[s] * x^3

    This is used to write constraints more compactly.

    Args:
        x: x variable, numeric, numpy array, or Pyomo component
        alpha: cubic parameters, numeric or Pyomo component
        s: optional segment index

    Returns:
        Pyomo expression, numpy array, or float
    """
    if s is None:
        return alpha[1] + alpha[2] * x + alpha[3] * x**2 + alpha[4] * x**3
    return alpha[s, 1] + alpha[s, 2] * x + alpha[s, 3] * x**2 + alpha[s, 4] * x**3


def _fx_cubic(x, alpha, s=None):
    """
    Cubic function first derivative:
        dy/dx = a2 + 2 * a3 * x + 3 * a4 * x^2

    Optionally if s is provided it is a segment index.
        dy/dx = a2[s] + 2 * a3[s] * x + 3 * a4[s] * x^2

    This is used to write constraints more compactly.

    Args:
        x: x variable, numeric, numpy array, or Pyomo component
        alpha: cubic parameters, numeric or Pyomo component
        s: optional segment index

    Returns:
        Pyomo expression, numpy array, or float
    """
    if s is None:
        return alpha[2] + 2 * alpha[3] * x + 3 * alpha[4] * x**2
    return alpha[s, 2] + 2 * alpha[s, 3] * x + 3 * alpha[s, 4] * x**2


def _fxx_cubic(x, alpha, s=None):
    """
    Cubic function second derivative:
        d2y/dx2 = 2 * a3 + 6 * a4 * x

    Optionally if s is provided it is a segment index.
        d2y/dx2 = 2 * a3[s] + 6 * a4[s] * x

    This is used to write constraints more compactly.

    Args:
        x: x variable, numeric, numpy array, or Pyomo component
        alpha: cubic parameters, numeric or Pyomo component
        s: optional segment index

    Returns:
        Pyomo expression, numpy array, or float
    """
    if s is None:
        return 2 * alpha[3] + 6 * alpha[4] * x
    return 2 * alpha[s, 3] + 6 * alpha[s, 4] * x


def _yxx_endpoint_is_zero_eqn(blk, s):
    """Rule which is used to add a constraint to set the second derivative
    at the first and last knot to zero.
    """
    if s == blk.seg_idx.last():
        j = blk.knt_idx.last()
    else:
        j = s
    return _fxx_cubic(blk.x[j], blk.alpha, s) == 0


class _increasing_constraint:
    def __init__(self, tol=0):
        """Rule functor for concave constraints on cubics."""
        self.tol = tol

    def __call__(self, blk, k):
        if k >= len(blk.knt_idx):
            s = k - 1
        else:
            s = k
        return _fx_cubic(blk.x[k], blk.alpha, s) >= self.tol


class _decreasing_constraint:
    def __init__(self, tol=0):
        """Rule functor for concave constraints on cubics."""
        self.tol = tol

    def __call__(self, blk, k):
        if k >= len(blk.knt_idx):
            s = k - 1
        else:
            s = k
        return _fx_cubic(blk.x[k], blk.alpha, s) <= -self.tol


class _convex_constraint:
    def __init__(self, tol=0):
        """Rule functor for concave constraints on cubics."""
        self.tol = tol

    def __call__(self, blk, k):
        if k >= len(blk.knt_idx):
            s = k - 1
        else:
            s = k
        return _fxx_cubic(blk.x[k], blk.alpha, s) >= self.tol


class _concave_constraint:
    def __init__(self, tol=0):
        """Rule functor for concave constraints on cubics."""
        self.tol = tol

    def __call__(self, blk, k):
        if k >= len(blk.knt_idx):
            s = k - 1
        else:
            s = k
        return _fxx_cubic(blk.x[k], blk.alpha, s) <= -self.tol


class CsplineParameters:
    def __init__(self, model=None, fptr=None):
        """Cubic spline parameters class.  This can be used to read and
        write parameters or calculate cubic spline function values and
        derivatives for testing.
        """
        if model is not None and fptr is not None:
            raise ValueError("Please specify at most one of model or fptr.")
        if model is not None:
            self.get_parameters_from_model(model)
        elif fptr is not None:
            self.get_parameters_from_file(fptr)
        else:
            self.knots = np.array([])
            self.a1 = np.array([])
            self.a2 = np.array([])
            self.a3 = np.array([])
            self.a4 = np.array([])

    @property
    def n_knots(self):
        """Number of knots"""
        return len(self.knots)

    @property
    def n_segments(self):
        """Number of segments"""
        return len(self.knots) - 1

    @property
    def valid(self):
        """Ensure that the number of knots and cubic parameters is valid"""
        return (
            len(self.a1) == self.n_segments
            and len(self.a2) == self.n_segments
            and len(self.a3) == self.n_segments
            and len(self.a4) == self.n_segments
        )

    def get_parameters_from_model(self, m):
        """Read parameters from a Pyomo model used to calculate them"""
        self.knots = [pyo.value(x) for x in m.x.values()]
        self.a1 = [None] * len(m.seg_idx)
        self.a2 = [None] * len(m.seg_idx)
        self.a3 = [None] * len(m.seg_idx)
        self.a4 = [None] * len(m.seg_idx)
        for s in m.seg_idx:
            self.a1[s - 1] = pyo.value(m.alpha[s, 1])
            self.a2[s - 1] = pyo.value(m.alpha[s, 2])
            self.a3[s - 1] = pyo.value(m.alpha[s, 3])
            self.a4[s - 1] = pyo.value(m.alpha[s, 4])
        self.knots = np.array(self.knots)
        self.a1 = np.array(self.a1)
        self.a2 = np.array(self.a2)
        self.a3 = np.array(self.a3)
        self.a4 = np.array(self.a4)

    def get_parameters_from_file(self, fptr):
        """Read parameters from a file"""
        # line 1: number of segments
        ns = int(fptr.readline())
        # Make param lists
        self.knots = [None] * (ns + 1)
        self.a1 = [None] * ns
        self.a2 = [None] * ns
        self.a3 = [None] * ns
        self.a4 = [None] * ns
        # Read params
        for i in range(ns + 1):
            self.knots[i] = float(fptr.readline())
        for a in [self.a1, self.a2, self.a3, self.a4]:
            for i in range(ns):
                a[i] = float(fptr.readline())
        self.knots = np.array(self.knots)
        self.a1 = np.array(self.a1)
        self.a2 = np.array(self.a2)
        self.a3 = np.array(self.a3)
        self.a4 = np.array(self.a4)

    def add_linear_extrapolation_segments(self):
        """Add a segment on the front and back of the cspline so that
        any extrapolation will be linear."""
        # We need to add a knot for a linear segment on the beginning and
        # end.  Since the first and last segment will be used for extrapolation,
        # and we want them to be linear, it doesn't really matter how far out
        # the knots are. To try to be roughly in line with the data scale we
        # just use the distance from the first to the last knot.
        dist = self.knots[-1] - self.knots[0]
        x = np.array([self.knots[0], self.knots[-1]])
        y = self.f(x)
        m = self.dfdx(x)
        b = y - m * x
        k = np.array([self.knots[0] - dist, self.knots[-1] + dist])

        self.knots = np.insert(self.knots, 0, k[0])
        self.a1 = np.insert(self.a1, 0, b[0])
        self.a2 = np.insert(self.a2, 0, m[0])
        self.a3 = np.insert(self.a3, 0, 0)
        self.a4 = np.insert(self.a4, 0, 0)
        self.knots = np.append(self.knots, k[1])
        self.a1 = np.append(self.a1, b[1])
        self.a2 = np.append(self.a2, m[1])
        self.a3 = np.append(self.a3, 0)
        self.a4 = np.append(self.a4, 0)

    def write_parameters(self, fptr):
        """Write parameters to a file"""
        assert self.valid
        fptr.write(f"{self.n_segments}\n")
        for l in [self.knots, self.a1, self.a2, self.a3, self.a4]:
            for x in l:
                fptr.write(f"{x}\n")

    def segment(self, x):
        """Get the spline segment containing x.

        Args:
            x: location, float or numpy array

        Returns:
            segment(s) containing x, if x is a numpy array a numpy
            array of integers is returned otherwise return an integer
        """
        s = np.searchsorted(self.knots, x)
        if isinstance(s, np.ndarray):
            # if x is before the first knot use the first segment
            s[s <= 0] = 1
            # if x is after the last knot use last segment to extrapolate
            s[s >= len(self.knots)] = len(self.knots) - 1
        else:
            if s <= 0:
                # if x is before the first knot use the first segment
                return 0
            if s >= len(self.knots):
                # if x is after the last knot use last segment to extrapolate
                return len(self.knots) - 2
        return s - 1

    def f(self, x):
        """Get f(x)

        Args:
            x: location, numpy array float

        Returns:
            f(x) numpy array if x is numpy array or float
        """
        s = self.segment(x)
        return self.a1[s] + self.a2[s] * x + self.a3[s] * x**2 + self.a4[s] * x**3

    def dfdx(self, x):
        """Get d/dx(f(x))

        Args:
            x: location, numpy array float

        Returns:
            df/dx numpy array if x is numpy array or float
        """
        s = self.segment(x)
        return self.a2[s] + 2 * self.a3[s] * x + 3 * self.a4[s] * x**2


@declare_custom_block(name="CubicParametersModel")
class CubicParametersModelData(BlockData):
    def __init__(self, *args, concrete=True, **kwargs):
        BlockData.__init__(self, *args, **kwargs)

    def add_model(
        self,
        x_data,
        y_data,
        x_knots=None,
        end_point_constraint=True,
        objective_form=False,
    ):
        """Add parameter model to the block.  By default this creates a square
        linear model, but optionally it can leave off the endpoint second
        derivative constraints and add an objective function for fitting data
        instead.  The purpose of the alternative least squares form is to allow
        the spline to be constrained in other ways that don't require a perfect
        data match. The knots don't need to be the same as the x data to allow,
        for example, additional segments for extrapolation. This is not the most
        computationally efficient way to calculate parameters, but since it is
        used to precalculate parameters, speed is not important.

        Args:
            x_data: sorted list of x data
            y_data: list of y data
            x_knots: optional sorted list of knots (default is to use x_data)
            end_point_constraint: if True add constraint that second derivative
                = 0 at endpoints (default=True)
            objective_form: if True write a least squares objective rather than
                constraints to match data (default=False)
            name: optional model name

        Returns:
            Pyomo ConcreteModel
        """
        n_data = len(x_data)
        assert n_data == len(y_data)
        if x_knots is None:
            n_knots = n_data
            x_knots = x_data
        else:
            n_knots = len(x_knots)

        self.knt_idx = pyo.RangeSet(n_knots)
        self.seg_idx = pyo.RangeSet(n_knots - 1)
        self.dat_idx = pyo.RangeSet(n_data)

        self.x_data = pyo.Param(
            self.dat_idx, initialize={i + 1: x for i, x in enumerate(x_data)}
        )
        self.y_data = pyo.Param(
            self.dat_idx, initialize={i + 1: x for i, x in enumerate(y_data)}
        )
        self.x = pyo.Param(
            self.knt_idx, initialize={i + 1: x for i, x in enumerate(x_knots)}
        )
        self.alpha = pyo.Var(self.seg_idx, {1, 2, 3, 4}, initialize=1)

        # f_s(x) = f_s+1(x)
        @self.Constraint(self.seg_idx)
        def y_eqn(blk, s):
            if s == self.seg_idx.last():
                return pyo.Constraint.Skip
            return _f_cubic(self.x[s + 1], self.alpha, s) == _f_cubic(
                self.x[s + 1], self.alpha, s + 1
            )

        # f'_s(x) = f'_s+1(x)
        @self.Constraint(self.seg_idx)
        def yx_eqn(blk, s):
            if s == self.seg_idx.last():
                return pyo.Constraint.Skip
            return _fx_cubic(self.x[s + 1], self.alpha, s) == _fx_cubic(
                self.x[s + 1], self.alpha, s + 1
            )

        # f"_s(x) = f"_s+1(x)
        @self.Constraint(self.seg_idx)
        def yxx_eqn(blk, s):
            if s == self.seg_idx.last():
                return pyo.Constraint.Skip
            return _fxx_cubic(self.x[s + 1], self.alpha, s) == _fxx_cubic(
                self.x[s + 1], self.alpha, s + 1
            )

        # Identify segments used to predict y_data at each x_data.  We use search in
        # instead of a dict lookup, since we don't want to require the data to be at
        # the knots, even though that is almost always the case.
        idx = np.searchsorted(x_knots, x_data)

        if end_point_constraint:
            self.add_endpoint_second_derivative_constraints()

        # Expression for difference between data and prediction
        @self.Expression(self.dat_idx)
        def ydiff(blk, d):
            s = idx[d - 1] + 1
            if s >= self.seg_idx.last():
                s = self.seg_idx.last()
            return self.y_data[d] - _f_cubic(self.x_data[d], self.alpha, s)

        if objective_form:
            # least squares objective
            self.obj = pyo.Objective(expr=sum(self.ydiff[d] ** 2 for d in self.dat_idx))
        else:

            @self.Constraint(self.dat_idx)
            def match_data(blk, d):
                return self.ydiff[d] == 0

    def add_endpoint_second_derivative_constraints(self):
        """Usually cubic splines use the endpoint constraints that the second
        derivative is zero.  This function adds those constraints to a model
        """
        self.second_derivative_at_endpoints = pyo.Constraint(
            [self.seg_idx.first(), self.seg_idx.last()], rule=_yxx_endpoint_is_zero_eqn
        )

    def add_decreasing_constraints(self, tol=0):
        """If the objective form of the parameter calculation is used, the
        data and the spline don't need to match exactly, and we can add
        constraints on the derivatives that they are negative at the knots.

        This doesn't necessarily mean the cubic spline function is always
        decreasing, since the segments are cubic, but you can either check the
        resulting curve or pair it with convex or concave constraints.
        """
        self.decreasing_ineq = pyo.Constraint(
            self.knt_idx, rule=_decreasing_constraint(tol)
        )

    def add_increasing_constraints(self, tol=0):
        """If the objective form of the parameter calculation is used, the
        data and the spline don't need to match exactly, and we can add
        constraints on the derivatives that they are positive at the knots.

        This doesn't necessarily mean the cubic spline function is always
        increasing, since the segments are cubic, but you can either check the
        resulting curve or pair it with convex or concave constraints.
        """
        self.increasing_ineq = pyo.Constraint(
            self.knt_idx, rule=_increasing_constraint(tol)
        )

    def add_concave_constraints(self, tol=0):
        """If the objective form of the parameter calculation is used, the
        data and the spline don't need to match exactly, and we can add
        constraints on the second derivatives that they are always negative.
        """
        self.concave_ineq = pyo.Constraint(self.knt_idx, rule=_concave_constraint(tol))

    def add_convex_constraints(self, tol=0):
        """If the objective form of the parameter calculation is used, the
        data and the spline don't need to match exactly, and we can add
        constraints on the second derivatives that they are always positive.
        """
        self.convex_ineq = pyo.Constraint(self.knt_idx, rule=_convex_constraint(tol))


def cubic_parameters_model(
    x_data,
    y_data,
    x_knots=None,
    end_point_constraint=True,
    objective_form=False,
    name="cubic spline parameters model",
):
    """Create a Pyomo model to calculate parameters for a cubic spline.  By default
    this creates a square linear model, but optionally it can leave off the endpoint
    second derivative constraints and add an objective function for fitting data
    instead.  The purpose of the alternative least squares form is to allow the spline
    to be constrained in other ways that don't require a perfect data match. The knots
    don't need to be the same as the x data to allow, for example, additional segments
    for extrapolation. This is not the most computationally efficient way to calculate
    parameters, but since it is used to precalculate parameters, speed is not important.

    Args:
        x_data: sorted list of x data
        y_data: list of y data
        x_knots: optional sorted list of knots (default is to use x_data)
        end_point_constraint: if True add constraint that second derivative = 0 at
            endpoints (default=True)
        objective_form: if True write a least squares objective rather than constraints
            to match data (default=False)
        name: optional model name

    Returns:
        Pyomo ConcreteModel
    """
    m = CubicParametersModel(name=name, concrete=True)
    m.add_model(
        x_data=x_data,
        y_data=y_data,
        x_knots=x_knots,
        end_point_constraint=end_point_constraint,
        objective_form=objective_form,
    )
    return m


def add_decreasing_constraints(m, tol=0):
    """If the objective form of the parameter calculation is used, the
    data and the spline don't need to match exactly, and we can add
    constraints on the derivatives that they are negative at the knots.

    This doesn't necessarily mean the cubic spline function is always
    decreasing, since the segments are cubic, but you can either check the
    resulting curve or pair it with convex or concave constraints.
    """
    m.add_decreasing_constraints(tol=tol)


def add_concave_constraints(m, tol=0):
    """If the objective form of the parameter calculation is used, the
    data and the spline don't need to match exactly, and we can add
    constraints on the second derivatives that they are always negative.
    """
    m.concave_ineq = pyo.Constraint(m.knt_idx, rule=_concave_constraint(tol))


def add_increasing_constraints(m, tol=0):
    """If the objective form of the parameter calculation is used, the
    data and the spline don't need to match exactly, and we can add
    constraints on the derivatives that they are positive at the knots.

    This doesn't necessarily mean the cubic spline function is always
    increasing, since the segments are cubic, but you can either check the
    resulting curve or pair it with convex or concave constraints.
    """
    m.add_increasing_constraints(tol=tol)


def add_convex_constraints(m, tol=0):
    """If the objective form of the parameter calculation is used, the
    data and the spline don't need to match exactly, and we can add
    constraints on the second derivatives that they are always positive.
    """
    m.convex_ineq = pyo.Constraint(m.knt_idx, rule=_convex_constraint(tol))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import subprocess

from pyomo.common import Executable
from pyomo.common.collections import Bunch
from pyomo.common.tempfiles import TempfileManager

from pyomo.opt.base import ProblemFormat, ResultsFormat
from pyomo.opt.base.solvers import _extract_version, SolverFactory
from pyomo.opt.results import SolverStatus
from pyomo.opt.solver import SystemCallSolver

import logging

logger = logging.getLogger('pyomo.solvers')


@SolverFactory.register('conopt', doc='The CONOPT NLP solver')
class CONOPT(SystemCallSolver):
    """
    An interface to the CONOPT optimizer that uses the AMPL Solver Library.
    """

    def __init__(self, **kwds):
        #
        # Call base constructor
        #
        kwds["type"] = "conopt"
        super(CONOPT, self).__init__(**kwds)
        #
        # Setup valid problem formats, and valid results for each problem format
        # Also set the default problem and results formats.
        #
        self._valid_problem_formats = [ProblemFormat.nl]
        self._valid_result_formats = {}
        self._valid_result_formats[ProblemFormat.nl] = [ResultsFormat.sol]
        self.set_problem_format(ProblemFormat.nl)

        # Note: Undefined capabilities default to 'None'
        self._capabilities = Bunch()
        self._capabilities.linear = True
        self._capabilities.integer = True
        self._capabilities.quadratic_objective = True
        self._capabilities.quadratic_constraint = True
        self._capabilities.sos1 = True
        self._capabilities.sos2 = True

    def _default_results_format(self, prob_format):
        return ResultsFormat.sol

    def _default_executable(self):
        executable = Executable("conopt")
        if not executable:
            logger.warning(
                "Could not locate the 'conopt' executable, "
                "which is required for solver %s" % self.name
            )
            self.enable = False
            return None
        return executable.path()

    def _get_version(self):
        """
        Returns a tuple describing the solver executable version.
        """
        solver_exec = self.executable()
        if solver_exec is None:
            return _extract_version('')
        results = subprocess.run(
            [solver_exec],
            timeout=self._version_timeout,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        return _extract_version(results.stdout)

    def create_command_line(self, executable, problem_files):
        assert self._problem_format == ProblemFormat.nl
        assert self._results_format == ResultsFormat.sol

        #
        # Define log file
        #
        if self._log_file is None:
            self._log_file = TempfileManager.create_tempfile(suffix="_conopt.log")

        fname = problem_files[0]
        if '.' in fname:
            tmp = fname.split('.')
            if len(tmp) > 2:
                fname = '.'.join(tmp[:-1])
            else:
                fname = tmp[0]
        self._soln_file = fname + ".sol"

        #
        # Define results file (since an external parser is used)
        #
        self._results_file = self._soln_file

        #
        # Define command line
        #
        env = os.environ.copy()
        #
        # Merge the PYOMO_AMPLFUNC (externals defined within
        # Pyomo/Pyomo) with any user-specified external function
        # libraries
        #
        if 'PYOMO_AMPLFUNC' in env:
            if 'AMPLFUNC' in env:
                env['AMPLFUNC'] += "\n" + env['PYOMO_AMPLFUNC']
            else:
                env['AMPLFUNC'] = env['PYOMO_AMPLFUNC']

        cmd = [executable, problem_files[0], '-AMPL']
        if self._timer:
            cmd.insert(0, self._timer)

        # GAH: I am going to re-add the code by Zev that passed options through
        # to the command line. I'm not sure what solvers this method of passing options
        # through the envstr variable works for, but it does not seem to work for cplex
        # or gurobi
        opt = []
        for key in self.options:
            if key == 'solver':
                continue
            if isinstance(self.options[key], str) and ' ' in self.options[key]:
                opt.append(key + "=\"" + str(self.options[key]) + "\"")
                cmd.append(str(key) + "=" + str(self.options[key]))
            elif key == 'subsolver':
                opt.append("solver=" + str(self.options[key]))
                cmd.append(str(key) + "=" + str(self.options[key]))
            else:
                opt.append(key + "=" + str(self.options[key]))
                cmd.append(str(key) + "=" + str(self.options[key]))

        envstr = "%s_options" % self.options.solver
        # Merge with any options coming in through the environment
        env[envstr] = " ".join(opt)

        return Bunch(cmd=cmd, log_file=self._log_file, env=env)

    def _postsolve(self):
        results = super(CONOPT, self)._postsolve()
        # Hack so that the locally optimal termination
        # condition for CONOPT does not trigger a warning.
        # For some reason it sets the solver_results_num to
        # 100 in this case, which is reserved for cases
        # where "optimal solution indicated, but error likely".
        if results.solver.id == 100 and 'Locally optimal' in results.solver.message:
            results.solver.status = SolverStatus.ok
        return results



# -*- coding: utf-8 -*-
# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import collections
import operator

from pyomo.common.deprecation import deprecated, relocated_module_attribute
from pyomo.common.errors import PyomoException, DeveloperError
from pyomo.common.numeric_types import (
    native_numeric_types,
    check_if_numeric_type,
    value,
)

# Note: There is a circular dependence between numeric_expr and this
# module: this module would like to reuse/build on
# numeric_expr._categorize_arg_type(), and numeric_expr.NumericValue
# needs to call the relational dispatchers here.  Instead of ensuring
# that one of the modules is fully declared before importing into the
# other, we will have BOTH modules assume that the other module has NOT
# been declared.
import pyomo.core.expr.numeric_expr as numeric_expr

from pyomo.core.expr.base import ExpressionBase
from pyomo.core.expr.boolean_value import BooleanValue
from pyomo.core.expr.expr_common import (
    ExpressionType,
    RELATIONAL_ARG_TYPE as ARG_TYPE,
    _binary_op_dispatcher_type_mapping,
)
from pyomo.core.expr.visitor import polynomial_degree
from pyomo.core.pyomoobject import PyomoObject

# -------------------------------------------------------
#
# Expression classes
#
# -------------------------------------------------------


def _categorize_relational_arg_type(arg):
    """Attempt to categorize an unknown object type into a RELATIONAL_ARG_TYPE

    Note that this can return the following types:
    - MUTABLE
    - ASNUMERIC
    - INVALID
    - NATIVE
    - PARAM
    - OTHER
    - INEQUALITY
    - INVALID_RELATIONAL
    """
    arg_type = numeric_expr._categorize_arg_type(arg)
    if arg_type is ARG_TYPE.INVALID:
        if isinstance(arg, PyomoObject):
            if isinstance(arg, InequalityExpression):
                arg_type = ARG_TYPE.INEQUALITY
            else:
                arg_type = ARG_TYPE.INVALID_RELATIONAL
    elif arg_type > ARG_TYPE.NATIVE and arg_type != ARG_TYPE.PARAM:
        arg_type = ARG_TYPE.OTHER
    return arg_type


def _categorize_relational_arg_types(*args):
    return tuple(_categorize_relational_arg_type(arg) for arg in args)


class RelationalExpression(ExpressionBase, BooleanValue):
    __slots__ = ('_args_',)

    EXPRESSION_SYSTEM = ExpressionType.RELATIONAL

    def __init__(self, args):
        self._args_ = args

    def __bool__(self):
        if self.is_constant():
            return bool(self())
        raise PyomoException("""
Cannot convert non-constant Pyomo expression (%s) to bool.
This error is usually caused by using a Var, unit, or mutable Param in a
Boolean context such as an "if" statement, or when checking container
membership or equality. For example,
    >>> m.x = Var()
    >>> if m.x >= 1:
    ...     pass
and
    >>> m.y = Var()
    >>> if m.y in [m.x, m.y]:
    ...     pass
would both cause this exception.""".strip() % (self,))

    @property
    def args(self):
        """
        Return the child nodes

        Returns: Either a list or tuple (depending on the node storage
            model) containing only the child nodes of this node
        """
        return self._args_[: self.nargs()]

    @deprecated(
        "is_relational() is deprecated in favor of "
        "is_expression_type(ExpressionType.RELATIONAL)",
        version='6.4.3',
    )
    def is_relational(self):
        return self.is_expression_type(ExpressionType.RELATIONAL)

    def is_potentially_variable(self):
        return any(
            arg.__class__ not in native_numeric_types and arg.is_potentially_variable()
            for arg in self._args_
        )

    def polynomial_degree(self):
        """
        Return the polynomial degree of the expression.

        Returns:
            A non-negative integer that is the polynomial
            degree if the expression is polynomial, or :const:`None` otherwise.
        """
        return polynomial_degree(self)

    def _compute_polynomial_degree(self, result):
        # NB: We can't use max() here because None (non-polynomial)
        # overrides a numeric value (and max() just ignores it)
        ans = 0
        for x in result:
            if x is None:
                return None
            elif ans < x:
                ans = x
        return ans

    def __eq__(self, other):
        """
        Equal to operator

        This method is called when Python processes statements of the form::

            self == other
            other == self
        """
        return _eq_dispatcher[self.__class__, other.__class__](self, other)

    def __lt__(self, other):
        """
        Less than operator

        This method is called when Python processes statements of the form::

            self < other
            other > self
        """
        return _lt_dispatcher[self.__class__, other.__class__](self, other)

    def __gt__(self, other):
        """
        Greater than operator

        This method is called when Python processes statements of the form::

            self > other
            other < self
        """
        return _lt_dispatcher[other.__class__, self.__class__](other, self)

    def __le__(self, other):
        """
        Less than or equal operator

        This method is called when Python processes statements of the form::

            self <= other
            other >= self
        """
        return _le_dispatcher[self.__class__, other.__class__](self, other)

    def __ge__(self, other):
        """
        Greater than or equal operator

        This method is called when Python processes statements of the form::

            self >= other
            other <= self
        """
        return _le_dispatcher[other.__class__, self.__class__](other, self)


class RangedExpression(RelationalExpression):
    """
    Ranged expressions, which define relations with a lower and upper bound::

        x < y < z
        x <= y <= z

    args:
        args (tuple): child nodes
        strict (tuple): flags that indicate whether the inequalities are strict
    """

    __slots__ = ('_strict',)
    PRECEDENCE = 9

    # Shared tuples for the most common RangedExpression objects encountered
    # in math programming.  Creating a single (shared) tuple saves memory
    STRICT = {
        False: (False, False),
        True: (True, True),
        (True, True): (True, True),
        (False, False): (False, False),
        (True, False): (True, False),
        (False, True): (False, True),
    }

    def __init__(self, args, strict):
        super(RangedExpression, self).__init__(args)
        self._strict = RangedExpression.STRICT[strict]

    def nargs(self):
        return 3

    def create_node_with_local_data(self, args):
        return self.__class__(args, self._strict)

    def _apply_operation(self, result):
        _l, _b, _r = result
        if not self._strict[0]:
            if not self._strict[1]:
                return _l <= _b and _b <= _r
            else:
                return _l <= _b and _b < _r
        elif not self._strict[1]:
            return _l < _b and _b <= _r
        else:
            return _l < _b and _b < _r

    def _to_string(self, values, verbose, smap):
        return "%s  %s  %s  %s  %s" % (
            values[0],
            "<="[: 2 - self._strict[0]],
            values[1],
            "<="[: 2 - self._strict[1]],
            values[2],
        )

    @property
    def strict(self):
        return self._strict


class InequalityExpression(RelationalExpression):
    """
    Inequality expressions, which define less-than or
    less-than-or-equal relations::

        x < y
        x <= y

    args:
        args (tuple): child nodes
        strict (bool): a flag that indicates whether the inequality is strict
    """

    __slots__ = ('_strict',)
    PRECEDENCE = 9

    def __init__(self, args, strict):
        super().__init__(args)
        self._strict = strict

    def nargs(self):
        return 2

    def create_node_with_local_data(self, args):
        return self.__class__(args, self._strict)

    def _apply_operation(self, result):
        _l, _r = result
        if self._strict:
            return _l < _r
        return _l <= _r

    def _to_string(self, values, verbose, smap):
        return "%s  %s  %s" % (values[0], "<="[: 2 - self._strict], values[1])

    @property
    def strict(self):
        return self._strict


class TrivialRelationalExpression(InequalityExpression):
    """A trivial relational expression

    Note that an inequality is sufficient to induce infeasibility and is
    simpler to relax (in the Big-M sense) than an equality.

    Note that we do not want to provide a trivial equality constraint as
    that can confuse solvers like Ipopt into believing that the model
    has fewer degrees of freedom than it actually has.

    """

    __slots__ = ('_name',)
    singleton = {}

    def __new__(cls, name, args):
        if name not in cls.singleton:
            cls.singleton[name] = super().__new__(cls)
            super().__init__(cls.singleton[name], args, False)
            cls.singleton[name]._name = name
        return cls.singleton[name]

    def __init__(self, name, args):
        # note that the meat of __init__ is called as part of __new__ above.
        assert args == self.args

    def __deepcopy__(self, memo):
        # Prevent deepcopy from duplicating this object
        return self

    def __reduce__(self):
        return self.__class__, (self._name, self._args_)

    def _to_string(self, values, verbose, smap):
        return self._name


def inequality(lower=None, body=None, upper=None, strict=False):
    """
    A utility function that can be used to declare inequality and
    ranged inequality expressions.  The expression::

        inequality(2, model.x)

    is equivalent to the expression::

        2 <= model.x

    The expression::

        inequality(2, model.x, 3)

    is equivalent to the expression::

        2 <= model.x <= 3

    .. note:: This ranged inequality syntax is deprecated in Pyomo.
        This function provides a mechanism for expressing
        ranged inequalities without chained inequalities.

    args:
        lower: an expression defines a lower bound
        body: an expression defines the body of a ranged constraint
        upper: an expression defines an upper bound
        strict (bool): A boolean value that indicates whether the inequality
            is strict.  Default is :const:`False`.

    Returns:
        A relational expression.  The expression is an inequality
        if any of the values :attr:`lower`, :attr:`body` or
        :attr:`upper` is :const:`None`.  Otherwise, the expression
        is a ranged inequality.
    """
    if lower is None:
        if body is None or upper is None:
            raise ValueError("Invalid inequality expression.")
        return InequalityExpression((body, upper), strict)
    if body is None:
        if lower is None or upper is None:
            raise ValueError("Invalid inequality expression.")
        return InequalityExpression((lower, upper), strict)
    if upper is None:
        return InequalityExpression((lower, body), strict)
    return RangedExpression((lower, body, upper), strict)


class EqualityExpression(RelationalExpression):
    """
    Equality expression::

        x == y
    """

    __slots__ = ()
    PRECEDENCE = 9

    def nargs(self):
        return 2

    def __bool__(self):
        lhs, rhs = self.args
        if lhs is rhs:
            return True
        return super().__bool__()

    def _apply_operation(self, result):
        _l, _r = result
        return _l == _r

    def _to_string(self, values, verbose, smap):
        return "%s  ==  %s" % (values[0], values[1])


class NotEqualExpression(RelationalExpression):
    """
    Not-equal expression::

        x != y
    """

    __slots__ = ()

    def nargs(self):
        return 2

    def __bool__(self):
        lhs, rhs = self.args
        if lhs is not rhs:
            return True
        return super().__bool__()

    def _apply_operation(self, result):
        _l, _r = result
        return _l != _r

    def _to_string(self, values, verbose, smap):
        return "%s  !=  %s" % (values[0], values[1])


def tuple_to_relational_expr(args):
    if len(args) == 2:
        return EqualityExpression(args)
    else:
        return inequality(*args)


def _invalid_relational(op_type, op_str, a, b):
    def no(*args):
        return False

    if getattr(a, 'is_expression_type', no)(ExpressionType.RELATIONAL):
        if getattr(b, 'is_expression_type', no)(ExpressionType.RELATIONAL):
            msg = (
                f"Cannot create an {op_type} where both "
                "sub-expressions are relational expressions:"
            )
        else:
            msg = (
                f"Cannot create an {op_type} where one of the "
                "sub-expressions is a relational expression:"
            )
    elif getattr(b, 'is_expression_type', no)(ExpressionType.RELATIONAL):
        msg = (
            f"Cannot create an {op_type} where one of the "
            "sub-expressions is a relational expression:"
        )
    elif getattr(a, 'is_component_type', no)() and a.is_indexed():
        msg = (
            f"Argument for {op_type} is an indexed numeric "
            f"value specified without an index:\n\t{a.name}\nIs this "
            "value defined over an index that you did not specify?"
        )
    elif getattr(b, 'is_component_type', no)() and b.is_indexed():
        msg = (
            f"Argument for {op_type} is an indexed numeric "
            f"value specified without an index:\n\t{b.name}\nIs this "
            "value defined over an index that you did not specify?"
        )
    else:
        msg = "Attempting to use a non-numeric type in a numeric expression context:"
    raise TypeError(msg + f"\n    {a}\n    {{{op_str}}}\n    {b}")


def _eq_invalid(a, b):
    _invalid_relational('EqualityExpression', '==', a, b)


def _eq_native(a, b):
    return a == b


def _eq_expr(a, b):
    return EqualityExpression((a, b))


def _eq_param_param(a, b):
    if a.is_constant():
        a = a.value
        if b.is_constant():
            return a == b.value
    elif b.is_constant():
        b = b.value
    return EqualityExpression((a, b))


def _eq_param_any(a, b):
    if a.is_constant():
        return a.value == b
    return EqualityExpression((a, b))


def _eq_any_param(a, b):
    if b.is_constant():
        return a == b.value
    return EqualityExpression((a, b))


def _register_new_eq_handler(a, b):
    types = _categorize_relational_arg_types(a, b)
    # Retrieve the appropriate handler, record it in the main
    # _eq_dispatcher dict (so this method is not called a second time for
    # these types)
    _eq_dispatcher[a.__class__, b.__class__] = handler = _eq_type_handler_mapping[types]
    # Call the appropriate handler
    return handler(a, b)


_eq_dispatcher = collections.defaultdict(lambda: _register_new_eq_handler)
_eq_type_handler_mapping = _binary_op_dispatcher_type_mapping(
    _eq_dispatcher,
    {
        (ARG_TYPE.NATIVE, ARG_TYPE.NATIVE): _eq_native,
        (ARG_TYPE.NATIVE, ARG_TYPE.PARAM): _eq_any_param,
        (ARG_TYPE.NATIVE, ARG_TYPE.OTHER): _eq_expr,
        (ARG_TYPE.NATIVE, ARG_TYPE.INEQUALITY): _eq_invalid,
        (ARG_TYPE.NATIVE, ARG_TYPE.INVALID_RELATIONAL): _eq_invalid,
        (ARG_TYPE.PARAM, ARG_TYPE.NATIVE): _eq_param_any,
        (ARG_TYPE.PARAM, ARG_TYPE.PARAM): _eq_param_param,
        (ARG_TYPE.PARAM, ARG_TYPE.OTHER): _eq_param_any,
        (ARG_TYPE.PARAM, ARG_TYPE.INEQUALITY): _eq_invalid,
        (ARG_TYPE.PARAM, ARG_TYPE.INVALID_RELATIONAL): _eq_invalid,
        (ARG_TYPE.OTHER, ARG_TYPE.NATIVE): _eq_expr,
        (ARG_TYPE.OTHER, ARG_TYPE.PARAM): _eq_any_param,
        (ARG_TYPE.OTHER, ARG_TYPE.OTHER): _eq_expr,
        (ARG_TYPE.OTHER, ARG_TYPE.INEQUALITY): _eq_invalid,
        (ARG_TYPE.OTHER, ARG_TYPE.INVALID_RELATIONAL): _eq_invalid,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.NATIVE): _eq_invalid,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.PARAM): _eq_invalid,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.OTHER): _eq_invalid,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.INEQUALITY): _eq_invalid,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.INVALID_RELATIONAL): _eq_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.NATIVE): _eq_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.PARAM): _eq_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.OTHER): _eq_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.INEQUALITY): _eq_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.INVALID_RELATIONAL): _eq_invalid,
    },
    ARG_TYPE,
)


def _le_invalid(a, b):
    _invalid_relational('InequalityExpression', '<=', a, b)


def _le_native(a, b):
    return a <= b


def _le_expr(a, b):
    return InequalityExpression((a, b), False)


def _le_expr_ineq(a, b):
    return RangedExpression((a,) + b.args, (False, b._strict))


def _le_param_ineq(a, b):
    if a.is_constant():
        a = a.value
    return RangedExpression((a,) + b.args, (False, b._strict))


def _le_ineq_expr(a, b):
    return RangedExpression(a.args + (b,), (a._strict, False))


def _le_ineq_param(a, b):
    if b.is_constant():
        b = b.value
    return RangedExpression(a.args + (b,), (a._strict, False))


def _le_param_param(a, b):
    if a.is_constant():
        a = a.value
        if b.is_constant():
            return a <= b.value
    elif b.is_constant():
        b = b.value
    return InequalityExpression((a, b), False)


def _le_param_any(a, b):
    if a.is_constant():
        return a.value <= b
    return InequalityExpression((a, b), False)


def _le_any_param(a, b):
    if b.is_constant():
        return a <= b.value
    return InequalityExpression((a, b), False)


def _register_new_le_handler(a, b):
    types = _categorize_relational_arg_types(a, b)
    # Retrieve the appropriate handler, record it in the main
    # _le_dispatcher dict (so this method is not called a second time for
    # these types)
    _le_dispatcher[a.__class__, b.__class__] = handler = _le_type_handler_mapping[types]
    # Call the appropriate handler
    return handler(a, b)


_le_dispatcher = collections.defaultdict(lambda: _register_new_le_handler)
_le_type_handler_mapping = _binary_op_dispatcher_type_mapping(
    _le_dispatcher,
    {
        (ARG_TYPE.NATIVE, ARG_TYPE.NATIVE): _le_native,
        (ARG_TYPE.NATIVE, ARG_TYPE.PARAM): _le_any_param,
        (ARG_TYPE.NATIVE, ARG_TYPE.OTHER): _le_expr,
        (ARG_TYPE.NATIVE, ARG_TYPE.INEQUALITY): _le_expr_ineq,
        (ARG_TYPE.NATIVE, ARG_TYPE.INVALID_RELATIONAL): _le_invalid,
        (ARG_TYPE.PARAM, ARG_TYPE.NATIVE): _le_param_any,
        (ARG_TYPE.PARAM, ARG_TYPE.PARAM): _le_param_param,
        (ARG_TYPE.PARAM, ARG_TYPE.OTHER): _le_param_any,
        (ARG_TYPE.PARAM, ARG_TYPE.INEQUALITY): _le_param_ineq,
        (ARG_TYPE.PARAM, ARG_TYPE.INVALID_RELATIONAL): _le_invalid,
        (ARG_TYPE.OTHER, ARG_TYPE.NATIVE): _le_expr,
        (ARG_TYPE.OTHER, ARG_TYPE.PARAM): _le_any_param,
        (ARG_TYPE.OTHER, ARG_TYPE.OTHER): _le_expr,
        (ARG_TYPE.OTHER, ARG_TYPE.INEQUALITY): _le_expr_ineq,
        (ARG_TYPE.OTHER, ARG_TYPE.INVALID_RELATIONAL): _le_invalid,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.NATIVE): _le_ineq_expr,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.PARAM): _le_ineq_param,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.OTHER): _le_ineq_expr,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.INEQUALITY): _le_invalid,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.INVALID_RELATIONAL): _le_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.NATIVE): _le_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.PARAM): _le_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.OTHER): _le_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.INEQUALITY): _le_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.INVALID_RELATIONAL): _le_invalid,
    },
    ARG_TYPE,
)


def _lt_invalid(a, b):
    _invalid_relational('InequalityExpression', '<', a, b)


def _lt_native(a, b):
    return a < b


def _lt_expr(a, b):
    return InequalityExpression((a, b), True)


def _lt_expr_ineq(a, b):
    return RangedExpression((a,) + b.args, (True, b._strict))


def _lt_param_ineq(a, b):
    if a.is_constant():
        a = a.value
    return RangedExpression((a,) + b.args, (True, b._strict))


def _lt_ineq_expr(a, b):
    return RangedExpression(a.args + (b,), (a._strict, True))


def _lt_ineq_param(a, b):
    if b.is_constant():
        b = b.value
    return RangedExpression(a.args + (b,), (a._strict, True))


def _lt_param_param(a, b):
    if a.is_constant():
        a = a.value
        if b.is_constant():
            return a < b.value
    elif b.is_constant():
        b = b.value
    return InequalityExpression((a, b), True)


def _lt_param_any(a, b):
    if a.is_constant():
        return a.value < b
    return InequalityExpression((a, b), True)


def _lt_any_param(a, b):
    if b.is_constant():
        return a < b.value
    return InequalityExpression((a, b), True)


def _register_new_lt_handler(a, b):
    types = _categorize_relational_arg_types(a, b)
    # Retrieve the appropriate handler, record it in the main
    # _lt_dispatcher dict (so this method is not called a second time for
    # these types)
    _lt_dispatcher[a.__class__, b.__class__] = handler = _lt_type_handler_mapping[types]
    # Call the appropriate handler
    return handler(a, b)


_lt_dispatcher = collections.defaultdict(lambda: _register_new_lt_handler)
_lt_type_handler_mapping = _binary_op_dispatcher_type_mapping(
    _lt_dispatcher,
    {
        (ARG_TYPE.NATIVE, ARG_TYPE.NATIVE): _lt_native,
        (ARG_TYPE.NATIVE, ARG_TYPE.PARAM): _lt_any_param,
        (ARG_TYPE.NATIVE, ARG_TYPE.OTHER): _lt_expr,
        (ARG_TYPE.NATIVE, ARG_TYPE.INEQUALITY): _lt_expr_ineq,
        (ARG_TYPE.NATIVE, ARG_TYPE.INVALID_RELATIONAL): _lt_invalid,
        (ARG_TYPE.PARAM, ARG_TYPE.NATIVE): _lt_param_any,
        (ARG_TYPE.PARAM, ARG_TYPE.PARAM): _lt_param_param,
        (ARG_TYPE.PARAM, ARG_TYPE.OTHER): _lt_param_any,
        (ARG_TYPE.PARAM, ARG_TYPE.INEQUALITY): _lt_param_ineq,
        (ARG_TYPE.PARAM, ARG_TYPE.INVALID_RELATIONAL): _lt_invalid,
        (ARG_TYPE.OTHER, ARG_TYPE.NATIVE): _lt_expr,
        (ARG_TYPE.OTHER, ARG_TYPE.PARAM): _lt_any_param,
        (ARG_TYPE.OTHER, ARG_TYPE.OTHER): _lt_expr,
        (ARG_TYPE.OTHER, ARG_TYPE.INEQUALITY): _lt_expr_ineq,
        (ARG_TYPE.OTHER, ARG_TYPE.INVALID_RELATIONAL): _lt_invalid,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.NATIVE): _lt_ineq_expr,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.PARAM): _lt_ineq_param,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.OTHER): _lt_ineq_expr,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.INEQUALITY): _lt_invalid,
        (ARG_TYPE.INEQUALITY, ARG_TYPE.INVALID_RELATIONAL): _lt_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.NATIVE): _lt_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.PARAM): _lt_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.OTHER): _lt_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.INEQUALITY): _lt_invalid,
        (ARG_TYPE.INVALID_RELATIONAL, ARG_TYPE.INVALID_RELATIONAL): _lt_invalid,
    },
    ARG_TYPE,
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr.numvalue import value as pyo_value
from pyomo.contrib.mpc.data.dynamic_data_base import _is_iterable, _DynamicDataBase
from pyomo.contrib.mpc.data.get_cuid import get_indexed_cuid


class ScalarData(_DynamicDataBase):
    """
    An object to store scalar data associated with time-indexed
    variables.
    """

    def __init__(self, data, time_set=None, context=None):
        """
        Arguments:
        ----------
        data: dict or ComponentMap
            Maps variables, names, or CUIDs to lists of values

        """
        for key, val in data.items():
            if _is_iterable(val):
                raise TypeError(
                    "Value %s corresponding to key %s is not a scalar" % (val, key)
                )
        super().__init__(data, time_set=time_set, context=context)

    def to_serializable(self):
        """
        Convert to json-serializable object.

        """
        data = {str(cuid): pyo_value(val) for cuid, val in self._data.items()}
        return data



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Model Reference: "A Dynamic HIV-Transmission Model for Evaluating the Costs
# 	and Benefits of Vaccine Programs", D.M. Edwards, R.D. Shachter,
# 	and D.K. Owen 1998, Interfaces
#


from pyomo.environ import (
    ConcreteModel,
    Param,
    Var,
    Objective,
    Constraint,
    Set,
    Expression,
    Suffix,
    value,
    exp,
    TransformationFactory,
)
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.dae.simulator import Simulator
from pyomo.contrib.sensitivity_toolbox.sens import sensitivity_calculation


def create_model():
    m = ConcreteModel()

    m.tf = Param(initialize=20)
    m.t = ContinuousSet(bounds=(0, m.tf))
    m.i = Set(initialize=[0, 1, 2, 3, 4, 5], ordered=True)
    m.j = Set(initialize=[0, 1], ordered=True)
    m.ij = Set(
        initialize=[(0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1), (3, 0), (4, 0)],
        ordered=True,
    )

    # Set Parameters
    m.eps = Param(initialize=0.75, mutable=True)

    m.sig = Param(initialize=0.15, mutable=True)
    m.xi = Param(initialize=0.983, mutable=True)
    m.omeg = Param(initialize=1.0 / 10, mutable=True)

    m.Y0 = Param(initialize=55816.0, mutable=True)
    m.phi0 = Param(initialize=0.493, mutable=True)

    d = {}
    d[(0, 0)] = 0.0222
    d[(0, 1)] = 0.0222
    d[(1, 0)] = 1 / 7.1
    d[(1, 1)] = 1 / 7.1
    d[(2, 0)] = 1 / 8.1
    d[(2, 1)] = 1 / (8.1 + 5)
    d[(3, 0)] = 1 / 2.7
    d[(4, 0)] = 1 / 2.1
    m.dd = Param(m.ij, initialize=d, mutable=True)

    d_inv = {}
    d_inv[(1, 0)] = 7.1
    d_inv[(2, 0)] = 8.1
    d_inv[(3, 0)] = 2.7
    d_inv[(4, 0)] = 2.1
    m.dd_inv = Param(m.ij, initialize=d_inv, default=0, mutable=True)

    I = {}
    I[(0, 0)] = 0.9 * m.dd[(0, 0)] * m.Y0
    I[(1, 0)] = 0.04 * m.dd[(0, 0)] * m.Y0
    I[(2, 0)] = 0.04 * m.dd[(0, 0)] * m.Y0
    I[(3, 0)] = 0.02 * m.dd[(0, 0)] * m.Y0
    m.II = Param(m.ij, initialize=I, default=0, mutable=True)

    p = {}
    p[(4, 0)] = 0.667
    m.pp = Param(m.ij, initialize=p, default=2, mutable=True)

    b = {}
    b[(1, 0)] = 0.066
    b[(1, 1)] = 0.066
    b[(2, 0)] = 0.066
    b[(2, 1)] = 0.066 * (1 - 0.25)
    b[(3, 0)] = 0.147
    b[(4, 0)] = 0.147
    m.bb = Param(m.ij, initialize=b, default=0, mutable=True)

    eta00 = {}
    eta00[(0, 0)] = 0.505
    eta00[(0, 1)] = 0.505
    eta00[(1, 0)] = 0.505
    eta00[(1, 1)] = 0.505
    eta00[(2, 0)] = 0.307
    eta00[(2, 1)] = 0.4803
    eta00[(3, 0)] = 0.235
    eta00[(4, 0)] = 0.235
    m.eta00 = Param(m.ij, initialize=eta00, mutable=True)

    eta01 = {}
    eta01[(0, 0)] = 0.505
    eta01[(0, 1)] = 0.6287
    eta01[(1, 0)] = 0.505
    eta01[(1, 1)] = 0.6287
    eta01[(2, 0)] = 0.307
    eta01[(2, 1)] = 0.4803
    eta01[(3, 0)] = 0.235
    eta01[(4, 0)] = 0.235
    m.eta01 = Param(m.ij, initialize=eta01, mutable=True)

    m.kp = Param(initialize=1000.0, mutable=True)
    m.kt = Param(initialize=1000.0, mutable=True)
    m.rr = Param(initialize=0.05, mutable=True)

    c = {}
    c[(0, 0)] = 3307
    c[(0, 1)] = 3307
    c[(1, 0)] = 5467
    c[(1, 1)] = 5467
    c[(2, 0)] = 5467
    c[(2, 1)] = 5467
    c[(3, 0)] = 12586
    c[(4, 0)] = 35394
    m.cc = Param(m.ij, initialize=c, mutable=True)

    q = {}
    q[(0, 0)] = 1
    q[(0, 1)] = 1
    q[(1, 0)] = 1
    q[(1, 1)] = 1
    q[(2, 0)] = 0.83
    q[(2, 1)] = 0.83
    q[(3, 0)] = 0.42
    q[(4, 0)] = 0.17
    m.qq = Param(m.ij, initialize=q, mutable=True)

    m.aa = Param(initialize=0.0001, mutable=True)

    # Set Variables
    m.yy = Var(m.t, m.ij)
    m.L = Var(m.t)

    m.vp = Var(m.t, initialize=0.75, bounds=(0, 0.75))
    m.vt = Var(m.t, initialize=0.75, bounds=(0, 0.75))

    m.dyy = DerivativeVar(m.yy, wrt=m.t)
    m.dL = DerivativeVar(m.L, wrt=m.t)

    def CostFunc(m):
        return m.L[m.tf]

    m.cf = Objective(rule=CostFunc)

    def _initDistConst(m):
        return (m.phi0 * m.Y0) / sum(m.dd_inv[kk] for kk in m.ij)

    m.idc = Expression(rule=_initDistConst)

    m.yy[0, (0, 0)].fix(value((1 - m.phi0) * m.Y0))
    m.yy[0, (0, 1)].fix(0)
    m.yy[0, (1, 0)].fix(value(m.dd_inv[(1, 0)] * m.idc))
    m.yy[0, (1, 1)].fix(0)
    m.yy[0, (2, 0)].fix(value(m.dd_inv[(2, 0)] * m.idc))
    m.yy[0, (2, 1)].fix(0)
    m.yy[0, (3, 0)].fix(value(m.dd_inv[(3, 0)] * m.idc))
    m.yy[0, (4, 0)].fix(value(m.dd_inv[(4, 0)] * m.idc))
    m.L[0].fix(0)

    # ODEs
    def _yy00(m, t):
        return (
            sum(m.pp[kk] * m.yy[t, kk] for kk in m.ij) * m.dyy[t, (0, 0)]
            == sum(m.pp[kk] * m.yy[t, kk] for kk in m.ij)
            * (
                m.II[(0, 0)]
                - (m.vp[t] + m.dd[(0, 0)]) * m.yy[t, (0, 0)]
                + m.omeg * m.yy[t, (0, 1)]
            )
            - m.pp[(0, 0)]
            * sum(m.bb[kk] * m.eta00[kk] * m.pp[kk] * m.yy[t, kk] for kk in m.ij)
            * m.yy[t, (0, 0)]
        )

    m.yy00DiffCon = Constraint(m.t, rule=_yy00)

    def _yy01(m, t):
        return (
            sum(m.pp[kk] * m.yy[t, kk] for kk in m.ij) * m.dyy[t, (0, 1)]
            == sum(m.pp[kk] * m.yy[t, kk] for kk in m.ij)
            * (m.vp[t] * m.yy[t, (0, 0)] - (m.dd[(0, 0)] + m.omeg) * m.yy[t, (0, 1)])
            - m.pp[(0, 1)]
            * (1 - m.eps)
            * sum(m.bb[kk] * m.eta01[kk] * m.pp[kk] * m.yy[t, kk] for kk in m.ij)
            * m.yy[t, (0, 1)]
        )

    m.yy01DiffCon = Constraint(m.t, rule=_yy01)

    def _yy10(m, t):
        return (
            sum(m.pp[kk] * m.yy[t, kk] for kk in m.ij) * m.dyy[t, (1, 0)]
            == sum(m.pp[kk] * m.yy[t, kk] for kk in m.ij)
            * (
                m.II[(1, 0)]
                - ((m.sig * m.xi) + m.vp[t] + m.dd[(1, 0)] + m.dd[(0, 0)])
                * m.yy[t, (1, 0)]
                + m.omeg * m.yy[t, (1, 1)]
            )
            + m.pp[(0, 0)]
            * sum(m.bb[kk] * m.eta00[kk] * m.pp[kk] * m.yy[t, kk] for kk in m.ij)
            * m.yy[t, (0, 0)]
        )

    m.yy10DiffCon = Constraint(m.t, rule=_yy10)

    def _yy11(m, t):
        return (
            sum(m.pp[kk] * m.yy[t, kk] for kk in m.ij) * m.dyy[t, (1, 1)]
            == sum(m.pp[kk] * m.yy[t, kk] for kk in m.ij)
            * (
                m.vp[t] * m.yy[t, (1, 0)]
                - (m.omeg + (m.sig * m.xi) + m.dd[(1, 1)] + m.dd[(0, 0)])
                * m.yy[t, (1, 1)]
            )
            + m.pp[(0, 1)]
            * (1 - m.eps)
            * sum(m.bb[kk] * m.eta01[kk] * m.pp[kk] * m.yy[t, kk] for kk in m.ij)
            * m.yy[t, (0, 1)]
        )

    m.yy11DiffCon = Constraint(m.t, rule=_yy11)

    def _yy20(m, t):
        return (
            m.dyy[t, (2, 0)]
            == m.II[(2, 0)]
            + m.sig * m.xi * (m.yy[t, (1, 0)] + m.yy[t, (1, 1)])
            - (m.vt[t] + m.dd[(2, 0)] + m.dd[(0, 0)]) * m.yy[t, (2, 0)]
        )

    m.yy20DiffCon = Constraint(m.t, rule=_yy20)

    def _yy21(m, t):
        return (
            m.dyy[t, (2, 1)]
            == m.vt[t] * m.yy[t, (2, 0)]
            - (m.dd[(2, 1)] + m.dd[(0, 0)]) * m.yy[t, (2, 1)]
        )

    m.yy21DiffCon = Constraint(m.t, rule=_yy21)

    def _yy30(m, t):
        return (
            m.dyy[t, (3, 0)]
            == m.II[(3, 0)]
            + m.dd[(1, 0)] * m.yy[t, (1, 0)]
            + m.dd[(1, 1)] * m.yy[t, (1, 1)]
            + m.dd[(2, 0)] * m.yy[t, (2, 0)]
            + m.dd[(2, 1)] * m.yy[t, (2, 1)]
            - (m.dd[(3, 0)] + m.dd[(0, 0)]) * m.yy[t, (3, 0)]
        )

    m.yy30DiffCon = Constraint(m.t, rule=_yy30)

    def _yy40(m, t):
        return (
            m.dyy[t, (4, 0)]
            == m.dd[(3, 0)] * m.yy[t, (3, 0)]
            - (m.dd[(4, 0)] + m.dd[(0, 0)]) * m.yy[t, (4, 0)]
        )

    m.yy40DiffCon = Constraint(m.t, rule=_yy40)

    def _L(m, t):
        return m.dL[t] == exp(-m.rr * t) * (
            m.aa
            * (
                m.kp * m.vp[t] * (m.yy[t, (0, 0)] + m.yy[t, (1, 0)])
                + (m.kt * m.vt[t] * m.yy[t, (2, 0)])
                + sum(m.cc[kk] * m.yy[t, kk] for kk in m.ij)
            )
            - (1 - m.aa) * sum(m.qq[kk] * m.yy[t, kk] for kk in m.ij)
        )

    m.LDiffCon = Constraint(m.t, rule=_L)

    return m


def initialize_model(m, n_sim, n_nfe, n_ncp):
    vp_profile = {0: 0.75}
    vt_profile = {0: 0.75}

    m.u_input = Suffix(direction=Suffix.LOCAL)
    m.u_input[m.vp] = vp_profile
    m.u_input[m.vt] = vt_profile

    sim = Simulator(m, package='scipy')
    tsim, profiles = sim.simulate(numpoints=n_sim, varying_inputs=m.u_input)

    discretizer = TransformationFactory('dae.collocation')
    discretizer.apply_to(m, nfe=n_nfe, ncp=n_ncp, scheme='LAGRANGE-RADAU')

    sim.initialize_model()


if __name__ == '__main__':
    m = create_model()
    initialize_model(m, 10, 5, 1)

    m.epsDelta = Param(initialize=0.75001)

    q_del = {}
    q_del[(0, 0)] = 1.001
    q_del[(0, 1)] = 1.002
    q_del[(1, 0)] = 1.003
    q_del[(1, 1)] = 1.004
    q_del[(2, 0)] = 0.83001
    q_del[(2, 1)] = 0.83002
    q_del[(3, 0)] = 0.42001
    q_del[(4, 0)] = 0.17001
    m.qqDelta = Param(m.ij, initialize=q_del)

    m.aaDelta = Param(initialize=0.0001001)

    m_sipopt = sensitivity_calculation(
        'sipopt', m, [m.eps, m.qq, m.aa], [m.epsDelta, m.qqDelta, m.aaDelta], tee=True
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.kernel.dict_container import DictContainer
from pyomo.core.kernel.tuple_container import TupleContainer
from pyomo.core.kernel.list_container import ListContainer


def define_homogeneous_container_type(
    namespace, name, container_class, ctype, doc=None, use_slots=True
):
    """
    This function is designed to be called for the simple
    container implementations (DictContainer, TupleContainer,
    and ListContainer) as the container_class argument.

    When called using the globals() namespace within a module, it
    is equivalent to placing the following class definition
    within that module:

    .. code::

        class <name>(<container_class>):
            <doc>
            _ctype = <ctype>

            ### if ### <use_slots>
            __slots__ = ("_parent",
                         "_storage_key",
                         "_active",
                         "_data",
                         "__weakref__")
            ### fi ###

            def __init__(self, *args, **kwds):
                self._parent = None
                self._storage_key = None
                self._active = True
                super(<name>, self).__init__(*args, **kwds)

    """
    assert name not in namespace
    cls_dict = {}
    cls_dict['_ctype'] = ctype
    if use_slots:
        cls_dict['__slots__'] = (
            "_parent",
            "_storage_key",
            "_active",
            "_data",
            "__weakref__",
        )

    def _init(self, *args, **kwds):
        self._parent = None
        self._storage_key = None
        self._active = True
        container_class.__init__(self, *args, **kwds)

    cls_dict['__init__'] = _init
    cls_dict['__module__'] = namespace['__name__']
    if doc is not None:
        cls_dict['__doc__'] = doc
    namespace[name] = type(name, (container_class,), cls_dict)


def define_simple_containers(namespace, prefix, ctype, use_slots=True):
    """Use this function to define all three simple
    container definitions for a new object category type."""
    doc_ = "A %s-style container for objects with category type " + ctype.__name__
    for suffix, container_class in (
        ('tuple', TupleContainer),
        ('list', ListContainer),
        ('dict', DictContainer),
    ):
        doc = doc_ % (suffix,)
        define_homogeneous_container_type(
            namespace,
            prefix + "_" + suffix,
            container_class,
            ctype,
            doc=doc,
            use_slots=use_slots,
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import attempt_import
from pyomo.solvers.plugins.solvers.gurobi_direct import gurobipy
from pyomo.contrib.mindtpy.cut_generation import add_oa_cuts, add_no_good_cuts
from pyomo.contrib.mcpp.pyomo_mcpp import McCormick as mc, MCPP_Error
from pyomo.repn import generate_standard_repn
import pyomo.core.expr as EXPR
from math import copysign
from pyomo.contrib.mindtpy.util import (
    get_integer_solution,
    copy_var_list_values,
    set_var_valid_value,
)
from pyomo.contrib.gdpopt.util import get_main_elapsed_time, time_code
from pyomo.opt import TerminationCondition as tc
from pyomo.core import minimize, value
from pyomo.core.expr import identify_variables

cplex, cplex_available = attempt_import('cplex')


class LazyOACallback_cplex(
    cplex.callbacks.LazyConstraintCallback if cplex_available else object
):
    """Inherent class in CPLEX to call Lazy callback."""

    def copy_lazy_var_list_values(
        self, opt, from_list, to_list, config, skip_stale=False, skip_fixed=True
    ):
        """This function copies variable values from one list to another.
        Rounds to Binary/Integer if necessary.
        Sets to zero for NonNegativeReals if necessary.

        Parameters
        ----------
        opt : SolverFactory
            The cplex_persistent solver.
        from_list : list
            The variable list that provides the values to copy from.
        to_list : list
            The variable list that needs to set value.
        config : ConfigBlock
            The specific configurations for MindtPy.
        skip_stale : bool, optional
            Whether to skip the stale variables, by default False.
        skip_fixed : bool, optional
            Whether to skip the fixed variables, by default True.
        """
        for v_from, v_to in zip(from_list, to_list):
            if skip_stale and v_from.stale:
                continue  # Skip stale variable values.
            if skip_fixed and v_to.is_fixed():
                continue  # Skip fixed variables.
            v_val = self.get_values(opt._pyomo_var_to_solver_var_map[v_from])
            set_var_valid_value(
                v_to,
                v_val,
                config.integer_tolerance,
                config.zero_tolerance,
                ignore_integrality=False,
            )

    def add_lazy_oa_cuts(
        self,
        target_model,
        dual_values,
        mindtpy_solver,
        config,
        opt,
        linearize_active=True,
        linearize_violated=True,
    ):
        """Linearizes nonlinear constraints; add the OA cuts through CPLEX inherent function self.add()
        For nonconvex problems, turn on 'config.add_slack'. Slack variables will always be used for
        nonlinear equality constraints.

        Parameters
        ----------
        target_model : Pyomo model
            The MIP main problem.
        dual_values : list
            The value of the duals for each constraint.
        mindtpy_solver : object
            The mindtpy solver class.
        config : ConfigBlock
            The specific configurations for MindtPy.
        opt : SolverFactory
            The cplex_persistent solver.
        linearize_active : bool, optional
            Whether to linearize the active nonlinear constraints, by default True.
        linearize_violated : bool, optional
            Whether to linearize the violated nonlinear constraints, by default True.
        """
        config.logger.debug('Adding OA cuts')
        with time_code(mindtpy_solver.timing, 'OA cut generation'):
            for index, constr in enumerate(target_model.MindtPy_utils.constraint_list):
                if (
                    constr.body.polynomial_degree()
                    in mindtpy_solver.mip_constraint_polynomial_degree
                ):
                    continue

                constr_vars = list(identify_variables(constr.body))
                jacs = mindtpy_solver.jacobians

                # Equality constraint (makes the problem nonconvex)
                if (
                    constr.has_ub()
                    and constr.has_lb()
                    and value(constr.lower) == value(constr.upper)
                ):
                    sign_adjust = (
                        -1 if mindtpy_solver.objective_sense == minimize else 1
                    )
                    rhs = constr.lower

                    # Since CPLEX requires the lazy cuts in CPLEX type,
                    # we need to transform the pyomo expression into CPLEX expression.
                    pyomo_expr = copysign(1, sign_adjust * dual_values[index]) * (
                        sum(
                            value(jacs[constr][var]) * (var - value(var))
                            for var in EXPR.identify_variables(constr.body)
                        )
                        + value(constr.body)
                        - rhs
                    )
                    cplex_expr, _ = opt._get_expr_from_pyomo_expr(pyomo_expr)
                    cplex_rhs = -generate_standard_repn(pyomo_expr).constant
                    self.add(
                        constraint=cplex.SparsePair(
                            ind=cplex_expr.variables, val=cplex_expr.coefficients
                        ),
                        sense='L',
                        rhs=cplex_rhs,
                    )
                    if (
                        self.get_solution_source()
                        == cplex.callbacks.SolutionSource.mipstart_solution
                    ):
                        mindtpy_solver.mip_start_lazy_oa_cuts.append(
                            [
                                cplex.SparsePair(
                                    ind=cplex_expr.variables,
                                    val=cplex_expr.coefficients,
                                ),
                                'L',
                                cplex_rhs,
                            ]
                        )
                else:  # Inequality constraint (possibly two-sided)
                    if (
                        constr.has_ub()
                        and (
                            linearize_active
                            and abs(constr.uslack()) < config.zero_tolerance
                        )
                        or (linearize_violated and constr.uslack() < 0)
                        or (config.linearize_inactive and constr.uslack() > 0)
                    ) or (
                        'MindtPy_utils.objective_constr' in constr.name
                        and constr.has_ub()
                    ):
                        pyomo_expr = sum(
                            value(jacs[constr][var]) * (var - var.value)
                            for var in constr_vars
                        ) + value(constr.body)
                        cplex_rhs = -generate_standard_repn(pyomo_expr).constant
                        cplex_expr, _ = opt._get_expr_from_pyomo_expr(pyomo_expr)
                        self.add(
                            constraint=cplex.SparsePair(
                                ind=cplex_expr.variables, val=cplex_expr.coefficients
                            ),
                            sense='L',
                            rhs=value(constr.upper) + cplex_rhs,
                        )
                        if (
                            self.get_solution_source()
                            == cplex.callbacks.SolutionSource.mipstart_solution
                        ):
                            mindtpy_solver.mip_start_lazy_oa_cuts.append(
                                [
                                    cplex.SparsePair(
                                        ind=cplex_expr.variables,
                                        val=cplex_expr.coefficients,
                                    ),
                                    'L',
                                    value(constr.upper) + cplex_rhs,
                                ]
                            )
                    if (
                        constr.has_lb()
                        and (
                            linearize_active
                            and abs(constr.lslack()) < config.zero_tolerance
                        )
                        or (linearize_violated and constr.lslack() < 0)
                        or (config.linearize_inactive and constr.lslack() > 0)
                    ) or (
                        'MindtPy_utils.objective_constr' in constr.name
                        and constr.has_lb()
                    ):
                        pyomo_expr = sum(
                            value(jacs[constr][var])
                            * (
                                var
                                - self.get_values(opt._pyomo_var_to_solver_var_map[var])
                            )
                            for var in constr_vars
                        ) + value(constr.body)
                        cplex_rhs = -generate_standard_repn(pyomo_expr).constant
                        cplex_expr, _ = opt._get_expr_from_pyomo_expr(pyomo_expr)
                        self.add(
                            constraint=cplex.SparsePair(
                                ind=cplex_expr.variables, val=cplex_expr.coefficients
                            ),
                            sense='G',
                            rhs=value(constr.lower) + cplex_rhs,
                        )
                        if (
                            self.get_solution_source()
                            == cplex.callbacks.SolutionSource.mipstart_solution
                        ):
                            mindtpy_solver.mip_start_lazy_oa_cuts.append(
                                [
                                    cplex.SparsePair(
                                        ind=cplex_expr.variables,
                                        val=cplex_expr.coefficients,
                                    ),
                                    'G',
                                    value(constr.lower) + cplex_rhs,
                                ]
                            )

    def add_lazy_affine_cuts(self, mindtpy_solver, config, opt):
        """Adds affine cuts using MCPP.

        Add affine cuts through CPLEX inherent function self.add().

        Parameters
        ----------
        mindtpy_solver : object
            The mindtpy solver class.
        config : ConfigBlock
            The specific configurations for MindtPy.
        opt : SolverFactory
            The cplex_persistent solver.
        """
        with time_code(mindtpy_solver.timing, 'Affine cut generation'):
            m = mindtpy_solver.mip
            config.logger.debug('Adding affine cuts')
            counter = 0

            for constr in m.MindtPy_utils.nonlinear_constraint_list:
                vars_in_constr = list(identify_variables(constr.body))
                if any(var.value is None for var in vars_in_constr):
                    continue  # a variable has no values

                # mcpp stuff
                try:
                    mc_eqn = mc(constr.body)
                except MCPP_Error as e:
                    config.logger.error(e, exc_info=True)
                    config.logger.debug(
                        'Skipping constraint %s due to MCPP error' % (constr.name)
                    )
                    continue  # skip to the next constraint
                ccSlope = mc_eqn.subcc()
                cvSlope = mc_eqn.subcv()
                ccStart = mc_eqn.concave()
                cvStart = mc_eqn.convex()

                concave_cut_valid = True
                convex_cut_valid = True
                for var in vars_in_constr:
                    if not var.fixed:
                        if ccSlope[var] == float('nan') or ccSlope[var] == float('inf'):
                            concave_cut_valid = False
                        if cvSlope[var] == float('nan') or cvSlope[var] == float('inf'):
                            convex_cut_valid = False
                if ccStart == float('nan') or ccStart == float('inf'):
                    concave_cut_valid = False
                if cvStart == float('nan') or cvStart == float('inf'):
                    convex_cut_valid = False
                # check if the value of ccSlope and cvSlope all equals zero. if so, we skip this.
                if not any(ccSlope.values()):
                    concave_cut_valid = False
                if not any(cvSlope.values()):
                    convex_cut_valid = False
                if not (concave_cut_valid or convex_cut_valid):
                    continue

                ub_int = (
                    min(value(constr.upper), mc_eqn.upper())
                    if constr.has_ub()
                    else mc_eqn.upper()
                )
                lb_int = (
                    max(value(constr.lower), mc_eqn.lower())
                    if constr.has_lb()
                    else mc_eqn.lower()
                )

                if concave_cut_valid:
                    pyomo_concave_cut = (
                        sum(
                            ccSlope[var] * (var - var.value)
                            for var in vars_in_constr
                            if not var.fixed
                        )
                        + ccStart
                    )
                    cplex_concave_rhs = generate_standard_repn(
                        pyomo_concave_cut
                    ).constant
                    cplex_concave_cut, _ = opt._get_expr_from_pyomo_expr(
                        pyomo_concave_cut
                    )
                    self.add(
                        constraint=cplex.SparsePair(
                            ind=cplex_concave_cut.variables,
                            val=cplex_concave_cut.coefficients,
                        ),
                        sense='G',
                        rhs=lb_int - cplex_concave_rhs,
                    )
                    counter += 1
                if convex_cut_valid:
                    pyomo_convex_cut = (
                        sum(
                            cvSlope[var] * (var - var.value)
                            for var in vars_in_constr
                            if not var.fixed
                        )
                        + cvStart
                    )
                    cplex_convex_rhs = generate_standard_repn(pyomo_convex_cut).constant
                    cplex_convex_cut, _ = opt._get_expr_from_pyomo_expr(
                        pyomo_convex_cut
                    )
                    self.add(
                        constraint=cplex.SparsePair(
                            ind=cplex_convex_cut.variables,
                            val=cplex_convex_cut.coefficients,
                        ),
                        sense='L',
                        rhs=ub_int - cplex_convex_rhs,
                    )
                    counter += 1

            config.logger.debug('Added %s affine cuts' % counter)

    def add_lazy_no_good_cuts(
        self, var_values, mindtpy_solver, config, opt, feasible=False
    ):
        """Adds no-good cuts.

        Add the no-good cuts through Cplex inherent function self.add().

        Parameters
        ----------
        var_values : list
            The variable values of the incumbent solution, used to generate the cut.
        mindtpy_solver : object
            The mindtpy solver class.
        config : ConfigBlock
            The specific configurations for MindtPy.
        opt : SolverFactory
            The cplex_persistent solver.
        feasible : bool, optional
            Whether the integer combination yields a feasible or infeasible NLP, by default False.

        Raises
        ------
        ValueError
            The value of binary variable is not 0 or 1.
        """
        if not config.add_no_good_cuts:
            return

        config.logger.debug('Adding no-good cuts')
        with time_code(mindtpy_solver.timing, 'No-good cut generation'):
            m = mindtpy_solver.mip
            MindtPy = m.MindtPy_utils
            int_tol = config.integer_tolerance

            binary_vars = [v for v in MindtPy.variable_list if v.is_binary()]

            # copy variable values over
            for var, val in zip(MindtPy.variable_list, var_values):
                if not var.is_binary():
                    continue
                # We don't want to trigger the reset of the global stale
                # indicator, so we will set this variable to be "stale",
                # knowing that set_value will switch it back to "not
                # stale"
                var.stale = True
                var.set_value(val, skip_validation=True)

            # check to make sure that binary variables are all 0 or 1
            for v in binary_vars:
                if value(abs(v - 1)) > int_tol and value(abs(v)) > int_tol:
                    raise ValueError(
                        'Binary {} = {} is not 0 or 1'.format(v.name, value(v))
                    )

            if not binary_vars:  # if no binary variables, skip
                return

            pyomo_no_good_cut = sum(
                1 - v for v in binary_vars if value(abs(v - 1)) <= int_tol
            ) + sum(v for v in binary_vars if value(abs(v)) <= int_tol)
            cplex_no_good_rhs = generate_standard_repn(pyomo_no_good_cut).constant
            cplex_no_good_cut, _ = opt._get_expr_from_pyomo_expr(pyomo_no_good_cut)

            self.add(
                constraint=cplex.SparsePair(
                    ind=cplex_no_good_cut.variables, val=cplex_no_good_cut.coefficients
                ),
                sense='G',
                rhs=1 - cplex_no_good_rhs,
            )

    def handle_lazy_main_feasible_solution(self, main_mip, mindtpy_solver, config, opt):
        """This function is called during the branch and bound of main mip, more
        exactly when a feasible solution is found and LazyCallback is activated.
        Copy the result to working model and update upper or lower bound.
        In LP-NLP, upper or lower bound are updated during solving the main problem.

        Parameters
        ----------
        main_mip : Pyomo model
            The MIP main problem.
        mindtpy_solver : object
            The mindtpy solver class.
        config : ConfigBlock
            The specific configurations for MindtPy.
        opt : SolverFactory
            The cplex_persistent solver.
        """
        # proceed. Just need integer values

        # this value copy is useful since we need to fix subproblem based on the solution of the main problem
        self.copy_lazy_var_list_values(
            opt,
            main_mip.MindtPy_utils.variable_list,
            mindtpy_solver.fixed_nlp.MindtPy_utils.variable_list,
            config,
            skip_fixed=False,
        )
        mindtpy_solver.update_dual_bound(self.get_best_objective_value())
        config.logger.info(
            mindtpy_solver.log_formatter.format(
                mindtpy_solver.mip_iter,
                'restrLP',
                self.get_objective_value(),
                mindtpy_solver.primal_bound,
                mindtpy_solver.dual_bound,
                mindtpy_solver.rel_gap,
                get_main_elapsed_time(mindtpy_solver.timing),
            )
        )

    def handle_lazy_subproblem_optimal(self, fixed_nlp, mindtpy_solver, config, opt):
        """This function copies the optimal solution of the fixed NLP subproblem to the MIP
        main problem(explanation see below), updates bound, adds OA and no-good cuts,
        stores incumbent solution if it has been improved.

        Parameters
        ----------
        fixed_nlp : Pyomo model
            Integer-variable-fixed NLP model.
        mindtpy_solver : object
            The mindtpy solver class.
        config : ConfigBlock
            The specific configurations for MindtPy.
        opt : SolverFactory
            The cplex_persistent solver.
        """
        if config.calculate_dual_at_solution:
            for c in fixed_nlp.tmp_duals:
                if fixed_nlp.dual.get(c, None) is None:
                    fixed_nlp.dual[c] = fixed_nlp.tmp_duals[c]
                elif (
                    config.nlp_solver == 'cyipopt'
                    and mindtpy_solver.objective_sense == minimize
                ):
                    # TODO: recover the opposite dual when cyipopt issue #2831 is solved.
                    fixed_nlp.dual[c] = -fixed_nlp.dual[c]
            dual_values = list(
                fixed_nlp.dual[c] for c in fixed_nlp.MindtPy_utils.constraint_list
            )
        else:
            dual_values = None
        main_objective = fixed_nlp.MindtPy_utils.objective_list[-1]
        mindtpy_solver.update_primal_bound(value(main_objective.expr))
        if mindtpy_solver.primal_bound_improved:
            mindtpy_solver.best_solution_found = fixed_nlp.clone()
            mindtpy_solver.best_solution_found_time = get_main_elapsed_time(
                mindtpy_solver.timing
            )
            if config.add_no_good_cuts or config.use_tabu_list:
                mindtpy_solver.stored_bound.update(
                    {mindtpy_solver.primal_bound: mindtpy_solver.dual_bound}
                )
        config.logger.info(
            mindtpy_solver.fixed_nlp_log_formatter.format(
                '*' if mindtpy_solver.primal_bound_improved else ' ',
                mindtpy_solver.nlp_iter,
                'Fixed NLP',
                value(main_objective.expr),
                mindtpy_solver.primal_bound,
                mindtpy_solver.dual_bound,
                mindtpy_solver.rel_gap,
                get_main_elapsed_time(mindtpy_solver.timing),
            )
        )

        # In OA algorithm, OA cuts are generated based on the solution of the subproblem
        # We need to first copy the value of variables from the subproblem and then add cuts
        # since value(constr.body), value(jacs[constr][var]), value(var) are used in self.add_lazy_oa_cuts()
        copy_var_list_values(
            fixed_nlp.MindtPy_utils.variable_list,
            mindtpy_solver.mip.MindtPy_utils.variable_list,
            config,
        )
        if config.strategy == 'OA':
            self.add_lazy_oa_cuts(
                mindtpy_solver.mip, dual_values, mindtpy_solver, config, opt
            )
            if config.add_regularization is not None:
                add_oa_cuts(
                    mindtpy_solver.mip,
                    dual_values,
                    mindtpy_solver.jacobians,
                    mindtpy_solver.objective_sense,
                    mindtpy_solver.mip_constraint_polynomial_degree,
                    mindtpy_solver.mip_iter,
                    config,
                    mindtpy_solver.timing,
                )
        elif config.strategy == 'GOA':
            self.add_lazy_affine_cuts(mindtpy_solver, config, opt)
        if config.add_no_good_cuts:
            var_values = list(v.value for v in fixed_nlp.MindtPy_utils.variable_list)
            self.add_lazy_no_good_cuts(var_values, mindtpy_solver, config, opt)

    def handle_lazy_subproblem_infeasible(self, fixed_nlp, mindtpy_solver, config, opt):
        """Solves feasibility NLP subproblem and adds cuts according to the specified strategy.

        Parameters
        ----------
        fixed_nlp : Pyomo model
            Integer-variable-fixed NLP model.
        mindtpy_solver : object
            The mindtpy solver class.
        config : ConfigBlock
            The specific configurations for MindtPy.
        opt : SolverFactory
            The cplex_persistent solver.
        """
        # TODO try something else? Reinitialize with different initial
        # value?
        config.logger.info('NLP subproblem was locally infeasible.')
        mindtpy_solver.nlp_infeasible_counter += 1
        if config.calculate_dual_at_solution:
            for c in fixed_nlp.MindtPy_utils.constraint_list:
                rhs = (0 if c.upper is None else c.upper) + (
                    0 if c.lower is None else c.lower
                )
                sign_adjust = 1 if c.upper is None else -1
                fixed_nlp.dual[c] = sign_adjust * max(
                    0, sign_adjust * (rhs - value(c.body))
                )
            dual_values = list(
                fixed_nlp.dual[c] for c in fixed_nlp.MindtPy_utils.constraint_list
            )
        else:
            dual_values = None

        config.logger.info('Solving feasibility problem')
        feas_subproblem, feas_subproblem_results = (
            mindtpy_solver.solve_feasibility_subproblem()
        )
        # In OA algorithm, OA cuts are generated based on the solution of the subproblem
        # We need to first copy the value of variables from the subproblem and then add cuts
        copy_var_list_values(
            feas_subproblem.MindtPy_utils.variable_list,
            mindtpy_solver.mip.MindtPy_utils.variable_list,
            config,
        )
        if config.strategy == 'OA':
            self.add_lazy_oa_cuts(
                mindtpy_solver.mip, dual_values, mindtpy_solver, config, opt
            )
            if config.add_regularization is not None:
                add_oa_cuts(
                    mindtpy_solver.mip,
                    dual_values,
                    mindtpy_solver.jacobians,
                    mindtpy_solver.objective_sense,
                    mindtpy_solver.mip_constraint_polynomial_degree,
                    mindtpy_solver.mip_iter,
                    config,
                    mindtpy_solver.timing,
                )
        elif config.strategy == 'GOA':
            self.add_lazy_affine_cuts(mindtpy_solver, config, opt)
        if config.add_no_good_cuts:
            var_values = list(v.value for v in fixed_nlp.MindtPy_utils.variable_list)
            self.add_lazy_no_good_cuts(var_values, mindtpy_solver, config, opt)

    def handle_lazy_subproblem_other_termination(
        self, fixed_nlp, termination_condition, mindtpy_solver, config
    ):
        """Handles the result of the latest iteration of solving the NLP subproblem given
        a solution that is neither optimal nor infeasible.

        Parameters
        ----------
        fixed_nlp : Pyomo model
            Integer-variable-fixed NLP model.
        termination_condition : Pyomo TerminationCondition
            The termination condition of the fixed NLP subproblem.
        mindtpy_solver : object
            The mindtpy solver class.
        config : ConfigBlock
            The specific configurations for MindtPy.

        Raises
        ------
        ValueError
            MindtPy unable to handle the termination condition of the fixed NLP subproblem.
        """
        if termination_condition is tc.maxIterations:
            # TODO try something else? Reinitialize with different initial value?
            config.logger.info(
                'NLP subproblem failed to converge within iteration limit.'
            )
            var_values = list(v.value for v in fixed_nlp.MindtPy_utils.variable_list)
        else:
            raise ValueError(
                'MindtPy unable to handle NLP subproblem termination '
                'condition of {}'.format(termination_condition)
            )

    def __call__(self):
        """This is an inherent function in LazyConstraintCallback in CPLEX.

        This function is called whenever an integer solution is found during the branch and bound process.
        """
        mindtpy_solver = self.mindtpy_solver
        config = self.config
        opt = self.opt
        main_mip = self.main_mip
        mindtpy_solver = self.mindtpy_solver

        # The lazy constraint callback may be invoked during MIP start processing. In that case get_solution_source returns mip_start_solution.
        # Reference: https://www.ibm.com/docs/en/icos/22.1.1?topic=SSSA5P_22.1.1/ilog.odms.cplex.help/refpythoncplex/html/cplex.callbacks.SolutionSource-class.htm
        # Another solution source is user_solution = 118, but it will not be encountered in LazyConstraintCallback.
        config.logger.info(
            "Solution source: {} (111 node_solution, 117 heuristic_solution, 119 mipstart_solution)".format(
                self.get_solution_source()
            )
        )

        # The solution found in MIP start process might be revisited in branch and bound.
        # Lazy constraints separated when processing a MIP start will be discarded after that MIP start has been processed.
        # This means that the callback may have to separate the same constraint again for the next MIP start or for a solution that is found later in the solution process.
        # https://www.ibm.com/docs/en/icos/22.1.1?topic=SSSA5P_22.1.1/ilog.odms.cplex.help/refpythoncplex/html/cplex.callbacks.LazyConstraintCallback-class.htm
        # For the MINLP3_simple example, all the solutions are obtained from mip_start (solution source). Therefore, it will not go to a branch and bound process.Cause an error output.
        if (
            self.get_solution_source()
            != cplex.callbacks.SolutionSource.mipstart_solution
            and len(mindtpy_solver.mip_start_lazy_oa_cuts) > 0
        ):
            for constraint, sense, rhs in mindtpy_solver.mip_start_lazy_oa_cuts:
                self.add(constraint, sense, rhs)
            mindtpy_solver.mip_start_lazy_oa_cuts = []

        if mindtpy_solver.should_terminate:
            # TODO: check the performance difference if we don't use self.abort() and let cplex terminate by itself.
            self.abort()
            return
        self.handle_lazy_main_feasible_solution(main_mip, mindtpy_solver, config, opt)
        if config.add_cuts_at_incumbent:
            self.copy_lazy_var_list_values(
                opt,
                main_mip.MindtPy_utils.variable_list,
                mindtpy_solver.mip.MindtPy_utils.variable_list,
                config,
            )
            if config.strategy == 'OA':
                # The solution obtained from mip start might be infeasible and even introduce a math domain error, like log(-1).
                try:
                    self.add_lazy_oa_cuts(
                        mindtpy_solver.mip, None, mindtpy_solver, config, opt
                    )
                except ValueError as e:
                    config.logger.error(e, exc_info=True)
                    config.logger.error(
                        "Usually this error is caused by the MIP start solution causing a math domain error. "
                        "We will skip it."
                    )
                    return

        # regularization is activated after the first feasible solution is found.
        if (
            config.add_regularization is not None
            and mindtpy_solver.best_solution_found is not None
        ):
            # The main problem might be unbounded, regularization is activated only when a valid bound is provided.
            if (
                not mindtpy_solver.dual_bound_improved
                and not mindtpy_solver.primal_bound_improved
            ):
                config.logger.debug(
                    'The bound and the best found solution have neither been improved.'
                    'We will skip solving the regularization problem and the Fixed-NLP subproblem'
                )
                mindtpy_solver.primal_bound_improved = False
                return
            if mindtpy_solver.dual_bound != mindtpy_solver.dual_bound_progress[0]:
                mindtpy_solver.add_regularization()
        if (
            abs(mindtpy_solver.primal_bound - mindtpy_solver.dual_bound)
            <= config.absolute_bound_tolerance
        ):
            config.logger.info(
                'MindtPy exiting on bound convergence. '
                '|Primal Bound: {} - Dual Bound: {}| <= (absolute tolerance {})  \n'.format(
                    mindtpy_solver.primal_bound,
                    mindtpy_solver.dual_bound,
                    config.absolute_bound_tolerance,
                )
            )
            mindtpy_solver.results.solver.termination_condition = tc.optimal
            # TODO: check the performance difference if we don't use self.abort() and let cplex terminate by itself.
            self.abort()
            return

        # check if the same integer combination is obtained.
        mindtpy_solver.curr_int_sol = get_integer_solution(
            mindtpy_solver.fixed_nlp, string_zero=True
        )

        if mindtpy_solver.curr_int_sol in set(mindtpy_solver.integer_list):
            config.logger.debug(
                'This integer combination has been explored. '
                'We will skip solving the Fixed-NLP subproblem.'
            )
            mindtpy_solver.primal_bound_improved = False
            if config.strategy == 'GOA':
                if config.add_no_good_cuts:
                    var_values = list(
                        v.value
                        for v in mindtpy_solver.working_model.MindtPy_utils.variable_list
                    )
                    self.add_lazy_no_good_cuts(var_values, mindtpy_solver, config, opt)
                return
            elif config.strategy == 'OA':
                return
        else:
            mindtpy_solver.integer_list.append(mindtpy_solver.curr_int_sol)

        # solve subproblem
        # Call the NLP pre-solve callback
        with time_code(mindtpy_solver.timing, 'Call before subproblem solve'):
            config.call_before_subproblem_solve(mindtpy_solver.fixed_nlp)
        # The constraint linearization happens in the handlers
        fixed_nlp, fixed_nlp_result = mindtpy_solver.solve_subproblem()
        # add oa cuts
        if fixed_nlp_result.solver.termination_condition in {
            tc.optimal,
            tc.locallyOptimal,
            tc.feasible,
        }:
            self.handle_lazy_subproblem_optimal(fixed_nlp, mindtpy_solver, config, opt)
            if (
                abs(mindtpy_solver.primal_bound - mindtpy_solver.dual_bound)
                <= config.absolute_bound_tolerance
            ):
                config.logger.info(
                    'MindtPy exiting on bound convergence. '
                    '|Primal Bound: {} - Dual Bound: {}| <= (absolute tolerance {})  \n'.format(
                        mindtpy_solver.primal_bound,
                        mindtpy_solver.dual_bound,
                        config.absolute_bound_tolerance,
                    )
                )
                mindtpy_solver.results.solver.termination_condition = tc.optimal
                return
        elif fixed_nlp_result.solver.termination_condition in {
            tc.infeasible,
            tc.noSolution,
        }:
            self.handle_lazy_subproblem_infeasible(
                fixed_nlp, mindtpy_solver, config, opt
            )
        else:
            self.handle_lazy_subproblem_other_termination(
                fixed_nlp,
                fixed_nlp_result.solver.termination_condition,
                mindtpy_solver,
                config,
            )


# Gurobi


def LazyOACallback_gurobi(cb_m, cb_opt, cb_where, mindtpy_solver, config):
    """This is a Gurobi callback function defined for LP/NLP based B&B algorithm.

    Parameters
    ----------
    cb_m : Pyomo model
        The MIP main problem.
    cb_opt : SolverFactory
        The gurobi_persistent solver.
    cb_where : int
        An enum member of gurobipy.GRB.Callback.
    mindtpy_solver : object
        The mindtpy solver class.
    config : ConfigBlock
        The specific configurations for MindtPy.
    """
    if cb_where == gurobipy.GRB.Callback.MIPSOL:
        # gurobipy.GRB.Callback.MIPSOL means that an integer solution is found during the branch and bound process
        if mindtpy_solver.should_terminate:
            cb_opt._solver_model.terminate()
            return
        cb_opt.cbGetSolution(vars=cb_m.MindtPy_utils.variable_list)
        handle_lazy_main_feasible_solution_gurobi(cb_m, cb_opt, mindtpy_solver, config)

        if config.add_cuts_at_incumbent:
            if config.strategy == 'OA':
                add_oa_cuts(
                    mindtpy_solver.mip,
                    None,
                    mindtpy_solver.jacobians,
                    mindtpy_solver.objective_sense,
                    mindtpy_solver.mip_constraint_polynomial_degree,
                    mindtpy_solver.mip_iter,
                    config,
                    mindtpy_solver.timing,
                    cb_opt=cb_opt,
                )

        # Regularization is activated after the first feasible solution is found.
        if (
            config.add_regularization is not None
            and mindtpy_solver.best_solution_found is not None
        ):
            # The main problem might be unbounded, regularization is activated only when a valid bound is provided.
            if (
                not mindtpy_solver.dual_bound_improved
                and not mindtpy_solver.primal_bound_improved
            ):
                config.logger.debug(
                    'The bound and the best found solution have neither been improved.'
                    'We will skip solving the regularization problem and the Fixed-NLP subproblem'
                )
                mindtpy_solver.primal_bound_improved = False
                return
            if mindtpy_solver.dual_bound != mindtpy_solver.dual_bound_progress[0]:
                mindtpy_solver.add_regularization()

        if mindtpy_solver.bounds_converged() or mindtpy_solver.reached_time_limit():
            cb_opt._solver_model.terminate()
            return

        # check if the same integer combination is obtained.
        mindtpy_solver.curr_int_sol = get_integer_solution(
            mindtpy_solver.fixed_nlp, string_zero=True
        )

        if mindtpy_solver.curr_int_sol in set(mindtpy_solver.integer_list):
            config.logger.debug(
                'This integer combination has been explored. '
                'We will skip solving the Fixed-NLP subproblem.'
            )
            mindtpy_solver.primal_bound_improved = False
            if config.strategy == 'GOA':
                if config.add_no_good_cuts:
                    var_values = list(
                        v.value
                        for v in mindtpy_solver.fixed_nlp.MindtPy_utils.variable_list
                    )
                    add_no_good_cuts(
                        mindtpy_solver.mip,
                        var_values,
                        config,
                        mindtpy_solver.timing,
                        mip_iter=mindtpy_solver.mip_iter,
                        cb_opt=cb_opt,
                    )
                return
            elif config.strategy == 'OA':
                # Refer to the official document of GUROBI.
                # Your callback should be prepared to cut off solutions that violate any of your lazy constraints, including those that have already been added. Node solutions will usually respect previously added lazy constraints, but not always.
                # https://www.gurobi.com/documentation/current/refman/cs_cb_addlazy.html
                # If this happens, MindtPy will look for the index of corresponding cuts, instead of solving the fixed-NLP again.
                begin_index, end_index = mindtpy_solver.integer_solution_to_cuts_index[
                    mindtpy_solver.curr_int_sol
                ]
                for ind in range(begin_index, end_index + 1):
                    cb_opt.cbLazy(mindtpy_solver.mip.MindtPy_utils.cuts.oa_cuts[ind])
                return
        else:
            mindtpy_solver.integer_list.append(mindtpy_solver.curr_int_sol)
            if config.strategy == 'OA':
                cut_ind = len(mindtpy_solver.mip.MindtPy_utils.cuts.oa_cuts)

        # solve subproblem
        # Call the NLP pre-solve callback
        with time_code(mindtpy_solver.timing, 'Call before subproblem solve'):
            config.call_before_subproblem_solve(mindtpy_solver.fixed_nlp)
        # The constraint linearization happens in the handlers
        fixed_nlp, fixed_nlp_result = mindtpy_solver.solve_subproblem()

        mindtpy_solver.handle_nlp_subproblem_tc(fixed_nlp, fixed_nlp_result, cb_opt)
        if config.strategy == 'OA':
            # store the cut index corresponding to current integer solution.
            mindtpy_solver.integer_solution_to_cuts_index[
                mindtpy_solver.curr_int_sol
            ] = [cut_ind + 1, len(mindtpy_solver.mip.MindtPy_utils.cuts.oa_cuts)]


def handle_lazy_main_feasible_solution_gurobi(cb_m, cb_opt, mindtpy_solver, config):
    """This function is called during the branch and bound of main MIP problem,
    more exactly when a feasible solution is found and LazyCallback is activated.

    Copy the solution to working model and update upper or lower bound.
    In LP-NLP, upper or lower bound are updated during solving the main problem.

    Parameters
    ----------
    cb_m : Pyomo model
        The MIP main problem.
    cb_opt : SolverFactory
        The gurobi_persistent solver.
    mindtpy_solver : object
        The mindtpy solver class.
    config : ConfigBlock
        The specific configurations for MindtPy.
    """
    # proceed. Just need integer values
    cb_opt.cbGetSolution(vars=cb_m.MindtPy_utils.variable_list)
    # this value copy is useful since we need to fix subproblem based on the solution of the main problem
    copy_var_list_values(
        cb_m.MindtPy_utils.variable_list,
        mindtpy_solver.fixed_nlp.MindtPy_utils.variable_list,
        config,
        skip_fixed=False,
    )
    copy_var_list_values(
        cb_m.MindtPy_utils.variable_list,
        mindtpy_solver.mip.MindtPy_utils.variable_list,
        config,
    )
    mindtpy_solver.update_dual_bound(cb_opt.cbGet(gurobipy.GRB.Callback.MIPSOL_OBJBND))
    config.logger.info(
        mindtpy_solver.log_formatter.format(
            mindtpy_solver.mip_iter,
            'restrLP',
            cb_opt.cbGet(gurobipy.GRB.Callback.MIPSOL_OBJ),
            mindtpy_solver.primal_bound,
            mindtpy_solver.dual_bound,
            mindtpy_solver.rel_gap,
            get_main_elapsed_time(mindtpy_solver.timing),
        )
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.common import unittest
from pyomo.contrib.observer.component_collector import collect_components_from_expr
from pyomo.common.collections import ComponentSet


class TestComponentCollector(unittest.TestCase):
    def test_nested_named_expressions(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.e1 = pyo.Expression(expr=m.x + m.y)
        m.e2 = pyo.Expression(expr=m.e1 + m.z)
        e = m.e2 * pyo.exp(m.e2)
        named_exprs, vars, params, external_funcs = collect_components_from_expr(e)
        self.assertEqual(len(named_exprs), 2)
        named_exprs = ComponentSet(named_exprs)
        self.assertIn(m.e1, named_exprs)
        self.assertIn(m.e2, named_exprs)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import pyomo.common.unittest as unittest
from pyomo.contrib.pynumero.dependencies import numpy as np, numpy_available

if not numpy_available:
    raise unittest.SkipTest('pynumero MA27 tests require numpy')
import ctypes
from pyomo.contrib.pynumero.linalg.ma27 import MA27Interface


@unittest.skipIf(not MA27Interface.available(), reason='MA27 not available')
class TestMA27Interface(unittest.TestCase):
    def test_get_cntl(self):
        ma27 = MA27Interface()
        self.assertEqual(ma27.get_icntl(1), 6)

        self.assertAlmostEqual(ma27.get_cntl(1), 1e-1)  # Numerical pivot threshold
        self.assertAlmostEqual(ma27.get_cntl(3), 0.0)  # Null pivot threshold

    def test_set_icntl(self):
        ma27 = MA27Interface()
        ma27.set_icntl(5, 4)  # Set output printing to max verbosity
        ma27.set_icntl(8, 1)  # Keep factors when we run out of space
        # (so MA27ED can be used)
        icntl5 = ma27.get_icntl(5)
        icntl8 = ma27.get_icntl(8)
        self.assertEqual(icntl5, 4)
        self.assertEqual(icntl8, 1)

        with self.assertRaisesRegex(TypeError, 'must be an integer'):
            ma27.set_icntl(1.0, 0)
        with self.assertRaisesRegex(IndexError, 'is out of range'):
            ma27.set_icntl(100, 0)
        with self.assertRaises(ctypes.ArgumentError):
            ma27.set_icntl(1, 0.0)

    def test_set_cntl(self):
        ma27 = MA27Interface()
        ma27.set_cntl(1, 1e-8)
        ma27.set_cntl(3, 1e-12)
        self.assertAlmostEqual(ma27.get_cntl(1), 1e-8)
        self.assertAlmostEqual(ma27.get_cntl(3), 1e-12)

    def test_do_symbolic_factorization(self):
        ma27 = MA27Interface()

        n = 5
        ne = 7
        irn = np.array([1, 1, 2, 2, 3, 3, 5], dtype=np.intc)
        icn = np.array([1, 2, 3, 5, 3, 4, 5], dtype=np.intc)
        # These arrays, copied out of HSL docs, contain Fortran indices.
        # Interfaces accept C indices as this is what I typically expect.
        irn = irn - 1
        icn = icn - 1

        bad_icn = np.array([1, 2, 3, 5, 3, 4], dtype=np.intc)
        # ^No need to update these indices

        ma27.do_symbolic_factorization(n, irn, icn)

        self.assertEqual(ma27.get_info(1), 0)
        self.assertEqual(ma27.get_info(5), 14)  # Min required num. integer words
        self.assertEqual(ma27.get_info(6), 20)  # Min required num. real words

        with self.assertRaisesRegex(AssertionError, 'Dimension mismatch'):
            ma27.do_symbolic_factorization(n, irn, bad_icn)

    def test_do_numeric_factorization(self):
        ma27 = MA27Interface()

        n = 5
        ne = 7
        irn = np.array([1, 1, 2, 2, 3, 3, 5], dtype=np.intc)
        icn = np.array([1, 2, 3, 5, 3, 4, 5], dtype=np.intc)
        irn = irn - 1
        icn = icn - 1
        ent = np.array([2.0, 3.0, 4.0, 6.0, 1.0, 5.0, 1.0], dtype=np.double)
        ma27.do_symbolic_factorization(n, irn, icn)

        status = ma27.do_numeric_factorization(irn, icn, n, ent)
        self.assertEqual(status, 0)

        expected_ent = [2.0, 3.0, 4.0, 6.0, 1.0, 5.0, 1.0]
        for i in range(ne):
            self.assertAlmostEqual(ent[i], expected_ent[i])

        self.assertEqual(ma27.get_info(15), 2)  # 2 negative eigenvalues
        self.assertEqual(ma27.get_info(14), 1)  # 1 2x2 pivot

        # Check that we can successfully perform another numeric factorization
        # with same symbolic factorization
        ent2 = np.array([1.5, 5.4, 1.2, 6.1, 4.2, 3.3, 2.0], dtype=np.double)
        status = ma27.do_numeric_factorization(irn, icn, n, ent2)
        self.assertEqual(ma27.get_info(15), 2)
        self.assertEqual(status, 0)

        with self.assertRaisesRegex(AssertionError, 'Dimension mismatch'):
            ma27.do_numeric_factorization(irn, icn, n + 1, ent)

        # Check that we can successfully perform another symbolic and
        # numeric factorization with the same ma27 struct
        #
        # n is still 5, ne has changed to 8.
        irn = np.array([1, 1, 2, 2, 3, 3, 5, 1], dtype=np.intc)
        icn = np.array([1, 2, 3, 5, 3, 4, 5, 5], dtype=np.intc)
        irn = irn - 1
        icn = icn - 1
        ent = np.array([2.0, 3.0, 4.0, 6.0, 1.0, 5.0, 1.0, 3.0], dtype=np.double)
        status = ma27.do_symbolic_factorization(n, irn, icn)
        self.assertEqual(status, 0)
        status = ma27.do_numeric_factorization(irn, icn, n, ent)
        self.assertEqual(status, 0)
        self.assertEqual(ma27.get_info(15), 3)

    def test_do_backsolve(self):
        ma27 = MA27Interface()

        n = 5
        ne = 7
        irn = np.array([1, 1, 2, 2, 3, 3, 5], dtype=np.intc)
        icn = np.array([1, 2, 3, 5, 3, 4, 5], dtype=np.intc)
        ent = np.array([2.0, 3.0, 4.0, 6.0, 1.0, 5.0, 1.0], dtype=np.double)
        irn = irn - 1
        icn = icn - 1
        rhs = np.array([8.0, 45.0, 31.0, 15.0, 17.0], dtype=np.double)
        status = ma27.do_symbolic_factorization(n, irn, icn)
        status = ma27.do_numeric_factorization(irn, icn, n, ent)
        sol = ma27.do_backsolve(rhs)

        expected_sol = [1, 2, 3, 4, 5]
        old_rhs = np.array([8.0, 45.0, 31.0, 15.0, 17.0])
        for i in range(n):
            self.assertAlmostEqual(sol[i], expected_sol[i])
            self.assertEqual(old_rhs[i], rhs[i])

        # Check that we can perform a numeric factorization with different ordering
        irn_mod = np.array([1, 2, 2, 1, 3, 3, 5], dtype=np.intc)
        icn_mod = np.array([2, 3, 5, 1, 3, 4, 5], dtype=np.intc)
        ent_mod = np.array([3.0, 4.0, 6.0, 2.0, 1.0, 5.0, 1.0], dtype=np.double)
        irn_mod -= 1
        icn_mod -= 1
        status = ma27.do_numeric_factorization(irn_mod, icn_mod, n, ent_mod)
        sol = ma27.do_backsolve(rhs)
        self.assertTrue(np.allclose(sol, np.array(expected_sol)))

        # Check that we can perform a numeric factorization with different lengths
        # due to extra zero entries
        irn_mod = irn_mod[1:]
        icn_mod = icn_mod[1:]
        ent_mod = ent_mod[1:]
        status = ma27.do_numeric_factorization(irn_mod, icn_mod, n, ent_mod)
        sol = ma27.do_backsolve(rhs)
        expected_sol = np.array([4.0, 1.91666666667, 3.0, 4.06666666667, 5.5])
        self.assertTrue(np.allclose(sol, expected_sol))


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import Counter
from pyomo.common.collections import ComponentSet
from pyomo.core.base import Constraint, Block
from pyomo.core.base.set import SetProduct


def index_warning(name, index):
    return 'WARNING: %s has no index %s' % (name, index)


def is_explicitly_indexed_by(comp, *sets, **kwargs):
    """
    Function for determining whether a pyomo component is indexed by a
    set or group of sets.

    Args:
        comp : Some Pyomo component, possibly indexed
        sets : Pyomo Sets to check indexing by
        expand_all_set_operators : Whether or not to expand all set operators
                                   in the subsets method

    Returns:
        A bool that is True if comp is directly indexed by every set in sets.
    """
    if not comp.is_indexed():
        return False
    for s in sets:
        if isinstance(s, SetProduct):
            msg = 'Checking for explicit indexing by a SetProduct is not supported'
            raise TypeError(msg)

    expand_all_set_operators = kwargs.pop('expand_all_set_operators', False)
    if kwargs:
        keys = kwargs.keys()
        raise ValueError('Unrecognized keyword arguments: %s' % str(keys))

    projected_subsets = comp.index_set().subsets(
        expand_all_set_operators=expand_all_set_operators
    )
    # Expanding all set operators here can be dangerous because it will not
    # distinguish between operators that contain their operands (e.g. union,
    # where you might consider the component to be considered indexed by
    # the operands) and operators that don't.
    # Ideally would like to check for containment by inclusion and containment
    # by product in one search of the set operators.
    subset_set = ComponentSet(projected_subsets)

    return all([_ in subset_set for _ in sets])


def is_in_block_indexed_by(comp, s, stop_at=None):
    """
    Function for determining whether a component is contained in a
    block that is indexed by a particular set.

    Args:
        comp : Component whose parent blocks are checked
        s : Set for which indices are checked
        stop_at : Block at which to stop searching if reached, regardless
                  of whether or not it is indexed by s

    Returns:
        Bool that is true if comp is contained in a block indexed by s
    """
    parent = comp.parent_block()

    # Stop when top-level block has been reached
    while parent is not None:
        # If we have reached our stopping point, quit.
        if parent is stop_at:
            return False

        # Look at the potentially-indexed block containing our component
        parent = parent.parent_component()
        # Check again for the stopping point in case an IndexedBlock was used
        if parent is stop_at:
            return False

        # Check potentially-indexed block for index s:
        if is_explicitly_indexed_by(parent, s):
            return True
        # Continue up the tree, checking the parent block of our
        # potentially-indexed block:
        else:
            parent = parent.parent_block()
    # Return False if top-level block was reached
    return False


def get_indices_of_projection(index_set, *sets):
    """
    We want to project certain sets out of our indexing set.
    We return the parameterization of this projection (the product
    of remaining sets) and a function to recover the original index
    from an element of the parameterization and coordinates of the
    sets projected out.
    """
    s_set = ComponentSet(sets)
    try:
        total_s_dim = sum([s.dimen for s in sets])
    except TypeError:
        msg = (
            'get_indices_of_projection does not support sets with '
            'dimen == None, including those with inconsistent dimen'
        )
        raise TypeError(msg)

    subset_set = ComponentSet(index_set.subsets())
    assert all(s in subset_set for s in sets)

    info = {}

    if isinstance(index_set, SetProduct):
        projection_sets = list(index_set.subsets())
        counter = Counter([id(_) for _ in projection_sets])
        for s in sets:
            if counter[id(s)] != 1:
                msg = 'Cannot omit sets that appear multiple times'
                raise ValueError(msg)
        # Need to know the location of each set within comp's index_set
        # location will map:
        #     location in comp's subsets() -> location in input sets
        location = {}
        # location should be well defined even for higher dimension sets
        # because this maps between lists of sets, not lists of indices
        other_ind_sets = []
        for ind_loc, ind_set in enumerate(projection_sets):
            found_set = False
            for s_loc, s_set in enumerate(sets):
                if ind_set is s_set:
                    location[ind_loc] = s_loc
                    found_set = True
                    break
            if not found_set:
                other_ind_sets.append(ind_set)
    else:
        # If index_set is not a SetProduct, only one set must have been
        # provided, so len(sets) == 1
        # Location in sets and in comp's indexing set are the same.
        location = {0: 0}
        other_ind_sets = []

    if index_set.dimen == total_s_dim:
        # comp indexed by all sets and having this dimension
        # is sufficient to know that comp is only indexed by
        # Sets in *sets

        # In this case, return the trivial set_except and index_getter

        # Problem: cannot construct location without a set tuple
        #          is that a problem with this syntax?
        #          Here len(newvals) should == 1
        info['set_except'] = [None]
        # index_getter returns an index corresponding to the values passed to
        # it, re-ordered according to order of indexing sets in component.
        info['index_getter'] = lambda incomplete_index, *newvals: (
            newvals[0]
            if len(newvals) <= 1
            else tuple([newvals[location[i]] for i in location])
        )
        return info

    # Now may assume other_ind_sets is nonempty.
    if len(other_ind_sets) == 1:
        set_except = other_ind_sets[0]
    elif len(other_ind_sets) >= 2:
        set_except = other_ind_sets[0].cross(*other_ind_sets[1:])
    else:
        raise ValueError('Did not expect this to happen')

    index_getter = lambda incomplete_index, *newvals: _complete_index(
        location, incomplete_index, *newvals
    )

    info['set_except'] = set_except
    info['index_getter'] = index_getter
    return info


def get_index_set_except(comp, *sets):
    """
    Function for getting indices of a component over a product of its
    indexing sets other than those specified. Indices for the specified
    sets can be used to construct indices of the proper dimension for the
    original component via the index_getter function.

    Args:
        comp : Component whose indexing sets are to be manipulated
        sets : Sets to omit from the set_except product

    Returns:
        A dictionary. Maps 'set_except' to a Pyomo Set or SetProduct
        of comp's index set, excluding those in sets. Maps
        'index_getter' to a function that returns an index of the
        proper dimension for comp, given an element of set_except
        and a value for each set excluded. These values must be provided
        in the same order their Sets were provided in the sets argument.
    """
    if not is_explicitly_indexed_by(comp, *sets):
        msg = (
            comp.name
            + ' is not indexed by at least one of '
            + str([s.name for s in sets])
        )
        raise ValueError(msg)

    return get_indices_of_projection(comp.index_set(), *sets)


def _complete_index(loc, index, *newvals):
    """
    Function for inserting new values into a partial index.
    Used by get_index_set_except function to construct the
    index_getter function for completing indices of a particular
    component with particular sets excluded.

    Args:
        loc : Dictionary mapping location in the new index to
              location in newvals
        index : Partial index
        newvals : New values to insert into index. Can be scalars
                  or tuples (for higher-dimension sets)

    Returns:
        An index (tuple) with values from newvals inserted in
        locations specified by loc
    """
    if type(index) is not tuple:
        index = (index,)
    keys = sorted(loc.keys())
    if len(keys) != len(newvals):
        raise ValueError('Wrong number of values to complete index')
    for i in sorted(loc.keys()):
        newval = newvals[loc[i]]
        if type(newval) is not tuple:
            newval = (newval,)
        index = index[0:i] + newval + index[i:]
    return index


def deactivate_model_at(b, cset, pts, allow_skip=True, suppress_warnings=False):
    """
    Finds any block or constraint in block b, indexed explicitly (and not
    implicitly) by cset, and deactivates it at points specified.
    Implicitly indexed components are excluded because one of their parent
    blocks will be deactivated, so deactivating them too would be redundant.

    Args:
        b : Block to search
        cset : ContinuousSet of interest
        pts : Value or list of values, in ContinuousSet, to deactivate at

    Returns:
        A dictionary mapping points in pts to lists of
        component data that have been deactivated there
    """
    if type(pts) is not list:
        pts = [pts]
    for pt in pts:
        if pt not in cset:
            msg = str(pt) + ' is not in ContinuousSet ' + cset.name
            raise ValueError(msg)
    deactivated = {pt: [] for pt in pts}

    visited = set()
    for comp in b.component_objects([Block, Constraint], active=True):
        # Record components that have been visited in case component_objects
        # contains duplicates (due to references)
        if id(comp) in visited:
            continue
        visited.add(id(comp))

        if is_explicitly_indexed_by(comp, cset) and not is_in_block_indexed_by(
            comp, cset
        ):
            info = get_index_set_except(comp, cset)
            non_cset_set = info['set_except']
            index_getter = info['index_getter']

            for non_cset_index in non_cset_set:
                for pt in pts:
                    index = index_getter(non_cset_index, pt)
                    try:
                        comp[index].deactivate()
                        deactivated[pt].append(comp[index])
                    except KeyError:
                        # except KeyError to allow Constraint/Block.Skip
                        if not suppress_warnings:
                            print(index_warning(comp.name, index))
                        if not allow_skip:
                            raise
                        continue

    return deactivated



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import json
from os.path import join, abspath, dirname
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.examples.semibatch.semibatch import SemiBatchExperiment


def main():

    # Data, list of dictionaries
    data = []
    file_dirname = dirname(abspath(str(__file__)))
    for exp_num in range(10):
        file_name = abspath(join(file_dirname, 'exp' + str(exp_num + 1) + '.out'))
        with open(file_name, 'r') as infile:
            d = json.load(infile)
            data.append(d)

    # Create an experiment list
    exp_list = []
    for i in range(len(data)):
        exp_list.append(SemiBatchExperiment(data[i]))

    # View one model
    # exp0_model = exp_list[0].get_labeled_model()
    # exp0_model.pprint()

    # Note, the model already includes a 'SecondStageCost' expression
    # for sum of squared error that will be used in parameter estimation

    pest = parmest.Estimator(exp_list)

    obj, theta = pest.theta_est()
    print(obj)
    print(theta)


if __name__ == '__main__':
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Test that EXPORT type Suffix components work with the NL writer
#

import os

import pyomo.common.unittest as unittest

from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager

from pyomo.opt import ProblemFormat
from pyomo.environ import (
    ConcreteModel,
    Suffix,
    Var,
    Objective,
    Constraint,
    SOSConstraint,
    sum_product,
)
from ..nl_diff import load_and_compare_nl_baseline

currdir = this_file_dir()


class SuffixTester:
    @classmethod
    def setUpClass(cls):
        cls.context = TempfileManager.new_context()
        cls.tempdir = cls.context.create_tempdir()

    @classmethod
    def tearDownClass(cls):
        cls.context.release()

    # test that EXPORT suffixes on variables,
    # constraints, objectives, and models
    # will end up in the NL file with integer tags
    def test_EXPORT_suffixes_int(self):
        model = ConcreteModel()
        model.junk = Suffix(direction=Suffix.EXPORT, datatype=Suffix.INT)
        model.junk_inactive = Suffix(direction=Suffix.EXPORT, datatype=Suffix.INT)

        model.x = Var()
        model.junk.set_value(model.x, 1)
        model.junk_inactive.set_value(model.x, 1)

        model.y = Var([1, 2], dense=True)
        model.junk.set_value(model.y, 2)
        model.junk_inactive.set_value(model.y, 2)

        model.obj = Objective(expr=model.x + sum_product(model.y))
        model.junk.set_value(model.obj, 3)
        model.junk_inactive.set_value(model.obj, 3)

        model.conx = Constraint(expr=model.x >= 1)
        model.junk.set_value(model.conx, 4)
        model.junk_inactive.set_value(model.conx, 4)

        model.cony = Constraint([1, 2], rule=lambda model, i: model.y[i] >= 1)
        model.junk.set_value(model.cony, 5)
        model.junk_inactive.set_value(model.cony, 5)

        model.junk.set_value(model, 6)
        model.junk_inactive.set_value(model, 6)

        # This one should NOT end up in the NL file
        model.junk_inactive.deactivate()

        _test = os.path.join(self.tempdir, "EXPORT_suffixes.test.nl")
        model.write(
            filename=_test,
            format=self.nl_version,
            io_options={"symbolic_solver_labels": False, "file_determinism": 1},
        )
        _base = os.path.join(currdir, "EXPORT_suffixes_int.baseline.nl")
        self.assertEqual(*load_and_compare_nl_baseline(_base, _test))

    # test that EXPORT suffixes on variables,
    # constraints, objectives, and models
    # will end up in the NL file with floating point tags
    def test_EXPORT_suffixes_float(self):
        model = ConcreteModel()
        model.junk = Suffix(direction=Suffix.EXPORT, datatype=Suffix.FLOAT)
        model.junk_inactive = Suffix(direction=Suffix.EXPORT, datatype=Suffix.FLOAT)

        model.x = Var()
        model.junk.set_value(model.x, 1)
        model.junk_inactive.set_value(model.x, 1)

        model.y = Var([1, 2], dense=True)
        model.junk.set_value(model.y, 2)
        model.junk_inactive.set_value(model.y, 2)

        model.obj = Objective(expr=model.x + sum_product(model.y))
        model.junk.set_value(model.obj, 3)
        model.junk_inactive.set_value(model.obj, 3)

        model.conx = Constraint(expr=model.x >= 1)
        model.junk.set_value(model.conx, 4)
        model.junk_inactive.set_value(model.conx, 4)

        model.cony = Constraint([1, 2], rule=lambda model, i: model.y[i] >= 1)
        model.junk.set_value(model.cony, 5)
        model.junk_inactive.set_value(model.cony, 5)

        model.junk.set_value(model, 6)
        model.junk_inactive.set_value(model, 6)

        # This one should NOT end up in the NL file
        model.junk_inactive.deactivate()

        _test = os.path.join(self.tempdir, "EXPORT_suffixes.test.nl")
        model.write(
            filename=_test,
            format=self.nl_version,
            io_options={"symbolic_solver_labels": False, "file_determinism": 1},
        )
        _base = os.path.join(currdir, "EXPORT_suffixes_float.baseline.nl")
        self.assertEqual(*load_and_compare_nl_baseline(_base, _test))

    # Test that user defined ref suffixes fail to
    # merge with those created from translating the SOSConstraint
    # component when variables get assigned duplicate values for ref
    def test_EXPORT_suffixes_with_SOSConstraint_duplicateref(self):
        model = ConcreteModel()
        model.ref = Suffix(direction=Suffix.EXPORT, datatype=Suffix.INT)
        model.y = Var([1, 2, 3])
        model.obj = Objective(expr=sum_product(model.y))

        # The NL writer will convert this constraint to ref and sosno
        # suffixes on model.y
        model.sos_con = SOSConstraint(var=model.y, index=[1, 2, 3], sos=1)

        for i, val in zip([1, 2, 3], [11, 12, 13]):
            model.ref.set_value(model.y[i], val)

        with self.assertRaisesRegex(
            RuntimeError,
            "NL file writer does not allow both manually "
            "declared 'ref' suffixes as well as SOSConstraint ",
        ):
            model.write(
                filename=os.path.join(self.tempdir, "junk.nl"),
                format=self.nl_version,
                io_options={"symbolic_solver_labels": False},
            )

    # Test that user defined sosno suffixes fail to
    # merge with those created from translating the SOSConstraint
    # component when variables get assigned duplicate values for sosno
    def test_EXPORT_suffixes_with_SOSConstraint_duplicatesosno(self):
        model = ConcreteModel()
        model.sosno = Suffix(direction=Suffix.EXPORT, datatype=Suffix.INT)
        model.y = Var([1, 2, 3])
        model.obj = Objective(expr=sum_product(model.y))

        # The NL writer will convert this constraint to ref and sosno
        # suffixes on model.y
        model.sos_con = SOSConstraint(var=model.y, index=[1, 2, 3], sos=1)

        for i in [1, 2, 3]:
            model.sosno.set_value(model.y[i], -1)

        with self.assertRaisesRegex(
            RuntimeError,
            "NL file writer does not allow both manually "
            "declared 'sosno' suffixes as well as SOSConstraint ",
        ):
            model.write(
                filename=os.path.join(self.tempdir, "junk.nl"),
                format=self.nl_version,
                io_options={"symbolic_solver_labels": False},
            )

    # Test that user defined sosno suffixes fail to
    # merge with those created from translating the SOSConstraint
    # component when variables get assigned duplicate values for sosno
    def test_EXPORT_suffixes_no_datatype(self):
        model = ConcreteModel()
        model.sosno = Suffix(direction=Suffix.EXPORT, datatype=None)
        model.y = Var([1, 2, 3])
        model.obj = Objective(expr=sum_product(model.y))

        # The NL writer will convert this constraint to ref and sosno
        # suffixes on model.y
        model.sos_con = SOSConstraint(var=model.y, index=[1, 2, 3], sos=1)

        for i in [1, 2, 3]:
            model.sosno.set_value(model.y[i], -1)

        with self.assertRaisesRegex(
            RuntimeError,
            "NL file writer does not allow both manually "
            "declared 'sosno' suffixes as well as SOSConstraint ",
        ):
            model.write(
                filename=os.path.join(self.tempdir, "junk.nl"),
                format=self.nl_version,
                io_options={"symbolic_solver_labels": False},
            )


class TestSuffix_nlv1(SuffixTester, unittest.TestCase):
    nl_version = 'nl_v1'


class TestSuffix_nlv2(SuffixTester, unittest.TestCase):
    nl_version = 'nl_v2'


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from .interfaces.model_interface import DynamicModelInterface
from .data.series_data import TimeSeriesData
from .data.interval_data import IntervalData
from .data.scalar_data import ScalarData
from .data.get_cuid import get_indexed_cuid
